package com.ethornium;

import com.ethornium.analysis.Analysis;
import com.ethornium.encoding.Contract;
import com.ethornium.encoding.InternalRepresentation;
import com.ethornium.util.Block;
import com.ethornium.util.ContractLexer;
import com.ethornium.util.GetBytecode;
import com.microsoft.z3.Context;
import com.microsoft.z3.Z3Exception;
import org.apache.commons.cli.*;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.*;
import java.math.BigInteger;
import java.nio.charset.Charset;
import java.util.*;
import java.util.concurrent.ExecutionException;

public class Main {
    private static final org.apache.commons.cli.Options options;
    private static final Logger LOGGER = LogManager.getLogger(Main.class);
    private static String INPUT_CONTRACT_PATH = "";
    private static String PATH_TO_DATADET = "";
    private static String ETHERSCANADRESSES = "/Users/ilya/Desktop/EtherScanTop100_09_02.txt";
    private static Options ethorniumOptions = new Options();
    private static ContractLexer.contractInfo contractInfo;
    private static Map<Integer, InternalRepresentation> parsedBytecodes;
    private static Vector<Integer> jumpDestinations;
    private static Integer maxPcInContract;
    private static String bytecode;
    private static Block block;
    private static Map<Contract, Integer> contracts;
    private static int boundary = 5; //boundary for loops (pre analysis)

    static {
        options = new org.apache.commons.cli.Options();
        options.addOption("t", false, "Execute test queries");
        options.addOption("g", false, "Global state dependency test");
        options.addOption("j", false, "Load contract from json, path to dataset");
        options.addOption("r", false, "Reentrancy check");
        options.addOption("p", false, "Skip pre-analysis step");
        options.addOption("s", true, "Memory size");
        options.addOption("d", false, "Print debug information");
        options.addOption("o", false, "Switch to online mode - get contracts from Ehterscan");
        options.addOption("a", true, "(Etherscan) Specifies the address of the contract, if none - fetches as many as possible");
        //options.addOption("q", true, "Time bound for the query (in seconds)");
        options.addOption("m", false, "Include memory inside the stack predicate");
    }

    private static String deleteHash(String bytecode){
        return bytecode.split("a165627a7a72305820")[0];
    }
    private static String ignore0x(String bytecode){
        if ((bytecode.charAt(0) == '0') && (bytecode.charAt(1) == 'x')){
            return bytecode.substring(2);
        }
        else
            return bytecode;
    }

    public static void analyse() throws IOException, ExecutionException, InterruptedException {
        for (Map.Entry<Contract, Integer> contract : contracts.entrySet()) {
            int contractID = contract.getValue();
            Analysis a = new Analysis(ethorniumOptions.GUARD, ethorniumOptions.test, ethorniumOptions.memSize, ethorniumOptions.reentrant, ethorniumOptions.globalDep, ethorniumOptions.memoryBound, ethorniumOptions.timeBound);
            LOGGER.info("Lexing " + contract.getKey().getName() + "...");
            bytecode = contract.getKey().getBytecode();
            bytecode = deleteHash(bytecode);
            bytecode = ignore0x(bytecode);
            lexContract(bytecode);
            LOGGER.info("Generating Horn Clauses...");
            a.generateHornClauses(parsedBytecodes, jumpDestinations, maxPcInContract, bytecode, contractID, contracts, boundary, ethorniumOptions.preAnalysis, ethorniumOptions.debug,
                    contract.getKey(), block);
            LOGGER.info("Analysing...");
            a.executeQueries(parsedBytecodes, contractID, contracts);
            LOGGER.info("Done!");
        }
    }

    public static void main(String args[]) throws Z3Exception, IOException, ExecutionException, InterruptedException, org.json.simple.parser.ParseException {
        LOGGER.info("Starting ethornium...");
        contracts = new LinkedHashMap<>();
        parseCommandLine(args);
        LOGGER.info("Time bound per query..." + Integer.toString(ethorniumOptions.timeBound) + " seconds");
        if (ethorniumOptions.memoryBound) {
            LOGGER.info("Memory is included in mState");
            LOGGER.info("Mempry size is..." + Integer.toString(ethorniumOptions.memSize) + " elements");
        }
        else{
            LOGGER.info("Separated memory predicate");
        }

        File blockf = new File("block.json");
        if (blockf.exists()){
            block = GetBytecode.getBlockInfoFromJsonFile();
        }
        else{
            block = GetBytecode.getBlockInfoFromEtherscan(ethorniumOptions.address);
        }

        if (!ethorniumOptions.online) {
            if (!ethorniumOptions.fromJson) {
                File inputFile = new File(INPUT_CONTRACT_PATH);
                LinkedHashSet<File> filesToProcess = new LinkedHashSet<>();
                if (!inputFile.exists()) {
                    throw new RuntimeException("Cannot find file or directory \"" + inputFile + "\"");
                }
                LOGGER.info("Collecting smart contract files to process...");
                if (inputFile.isDirectory()) {
                    getFilesInDir(inputFile, filesToProcess);
                } else if (inputFile.isFile()) {
                    filesToProcess.add(inputFile);
                }
                int contractID;
                for (File f : filesToProcess) {
                    if (!getContractIDs(f)) { // we have only one contract in the file
                        contractID = 0;
                        InputStream fis = new FileInputStream(f);
                        InputStreamReader isr = new InputStreamReader(fis, Charset.forName("UTF-8"));
                        BufferedReader br = new BufferedReader(isr);
                        String bytecode = br.readLine();
                        contracts.put(new Contract(f.getAbsolutePath(), contractID, bytecode), contractID);
                    }
                }
            }
            else{
                contracts.put(GetBytecode.getContractInfoFromJsonFile(INPUT_CONTRACT_PATH), 0);
            }
        }
        else {
            //GetBytecode.getTopContracts();
            //GetBytecode.makeDatabase();
            if (ethorniumOptions.address != null) {
                LOGGER.info("Getting from Ethrescan contract address: " + ethorniumOptions.address + "...");
                String bytecode = GetBytecode.getBytecodeFromEtherscan(ethorniumOptions.address);
                Contract contr = new Contract(ethorniumOptions.address, 0, bytecode);
                contr = GetBytecode.getContractData(ethorniumOptions.address, contr);
                contracts.put(contr, 0);
            }
            else{
                LOGGER.info("Getting MANY contracts from Ethrescan, file with addresses: " + ETHERSCANADRESSES + "...");
                File addresses = new File(ETHERSCANADRESSES);
                InputStream fis = new FileInputStream(addresses);
                InputStreamReader isr = new InputStreamReader(fis, Charset.forName("UTF-8"));
                BufferedReader br = new BufferedReader(isr);
                String address;
                int count = 0;
                while ((address = br.readLine()) != null) {
                    String bytecode = GetBytecode.getBytecodeFromEtherscan(address);
                    if (!bytecode.equals("0x")) {
                        contracts.put(new Contract(address, count, bytecode), count);
                        count++;
                    }
                    else {
                        LOGGER.debug(address + " has an empty bytecode!");
                    }
                }
            }
        }
        analyse();
    }

    private static boolean getContractIDs(File file) throws IOException {
        int contractID = 0;
        String line;
        boolean multipleContracts = false;
        InputStream fis = new FileInputStream(file);
        InputStreamReader isr = new InputStreamReader(fis, Charset.forName("UTF-8"));
        BufferedReader br = new BufferedReader(isr);
        while ((line = br.readLine()) != null) {
                if (line.charAt(0) == '%'){
                    String name = line.split("%")[1];
                    String bytecode = br.readLine();
                    contracts.put(new Contract(name, contractID, bytecode), contractID);
                    multipleContracts = true;
                    contractID++;
                }

        }
        return multipleContracts;
    }

    private static void lexContract(final String bytecode) throws IOException{
        contractInfo = ContractLexer.toOpcodes(bytecode, ethorniumOptions.debug);
        parsedBytecodes = contractInfo.getInternalByteCode();
        jumpDestinations = contractInfo.getJumpDestinations();
        maxPcInContract = Collections.max(parsedBytecodes.keySet());
        //maxPcInContract = parsedBytecodes.get(parsedBytecodes.size() - 1).getPc();
    }

    private static void parseCommandLine(String[] args) {
        CommandLineParser parser = new DefaultParser();
        CommandLine commandLine;
        try {
            commandLine = parser.parse(options, args);
        } catch (ParseException ex) {
            usage();
            return;
        }
        String[] otherArgs = commandLine.getArgs();
        Option[] clOptions = commandLine.getOptions();
        getOptionsDirective(commandLine, clOptions);

        if (!ethorniumOptions.online) {
            INPUT_CONTRACT_PATH = otherArgs[0];
        }
    }

    private static void usage() {
    }

    private static void getOptionsDirective(CommandLine commandLine, Option[] clOptions) {
        for (Option option : clOptions) {
            String opt = option.getOpt();
            switch (opt.charAt(0)) {
                case 't':
                    ethorniumOptions.test = true;
                    break;
                case 'r':
                    ethorniumOptions.reentrant = true;
                    break;
                case 'p':
                    ethorniumOptions.preAnalysis = false;
                    break;
                case 's':
                    ethorniumOptions.memSize = Integer.parseInt(commandLine.getOptionValue("s"));
                    break;
                case 'd':
                    ethorniumOptions.debug = true;
                    break;
                case 'o':
                    ethorniumOptions.online = true;
                    break;
                case 'a':
                    ethorniumOptions.address = commandLine.getOptionValue("a");
                    break;
                /*case 'q':
                    ethorniumOptions.timeBound = Integer.parseInt(commandLine.getOptionValue("q"));
                    break;*/
                case 'm':
                    ethorniumOptions.memoryBound = true;
                    break;
                case 'j':
                    ethorniumOptions.fromJson = true;
                    break;
                case 'g':
                    ethorniumOptions.globalDep = true;
                    break;
            }
        }
    }

    private static void getFilesInDir(File dir, Set<File> smartFiles) {
        File[] files = dir.listFiles();
        if (files != null) {
            for (File file : files) {
                if (file.isFile()) {
                    if (file.getName().endsWith(".txt")) {
                        smartFiles.add(file);
                    }
                } else if (file.isDirectory())
                    getFilesInDir(file.getAbsoluteFile(), smartFiles);
            }
        }
    }
}
