package com.ethornium.analysis;

import com.ethornium.encoding.*;
import com.ethornium.util.Block;
import com.ethornium.z3.Query;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Map;
import java.util.Vector;
import java.util.concurrent.ExecutionException;

public class Analysis {
    private final ArrayList<EFunc> mFuncs;
    private final ArrayList<EExpr> mRules;
    private final ArrayList<EExpr> mQueriesRe;
    private final ArrayList<EExpr> mQueriesDe;
    private final Pred pred;
    private final VarMap varMap;
    private final int GUARD;
    private boolean testQueries;
    private boolean reentrancy;
    private boolean dependancy;
    private final int memSize;
    private boolean memBound;
    private int timeBound;
    public Analysis(int GUARD, boolean testQueries, int memSize, boolean reentrancy, boolean dependancy, boolean memBound, int timeBound){
        this.GUARD = GUARD;
        mFuncs = new ArrayList<>();
        mRules = new ArrayList<>();
        mQueriesRe = new ArrayList<>();
        mQueriesDe = new ArrayList<>();
        this.pred = new Pred(mFuncs);
        this.varMap = new VarMap(GUARD, memSize);
        this.testQueries = testQueries;
        this.memSize = memSize;
        this.reentrancy = reentrancy;
        this.dependancy = dependancy;
        this.memBound = memBound;
        this.timeBound = timeBound;
    }

    public void generateHornClauses(final Map<Integer, InternalRepresentation> parsedBytecodes,
                                    final Vector<Integer> jumpDestinations, final Integer maxPcInContract, final String bytecode,
                                    final int contractID,
                                    Map<Contract, Integer> contracts, int boundary,
                                    boolean preAnalysis, boolean debug, Contract contract, Block block){

        if (preAnalysis){
            PreAnalysis pa = new PreAnalysis(
                parsedBytecodes, jumpDestinations,
                maxPcInContract, contractID,
                contracts, boundary, debug
            );
            pa.process(0);
        }

        InitializingRules ir = new InitializingRules(mRules, pred, jumpDestinations, maxPcInContract, bytecode,
                contractID,
                contracts, varMap, memBound);
        ir.process();

        HornClauseGenerator hcg = new HornClauseGenerator(mRules, mQueriesRe, mQueriesDe, pred, parsedBytecodes, jumpDestinations,
                maxPcInContract, contractID,
                contracts, varMap, debug, memSize, memBound, 0, contract, block, reentrancy, dependancy, true);
        hcg.GenerateHornClauses();

        hcg = new HornClauseGenerator(mRules, mQueriesRe, mQueriesDe, pred, parsedBytecodes, jumpDestinations,
                maxPcInContract, contractID,
                contracts, varMap, debug, memSize, memBound, 1, contract, block, reentrancy, dependancy, false);
        hcg.GenerateHornClauses();
    }

    public void executeQueries(final Map<Integer, InternalRepresentation> parsedBytecodes,
                               final int contractID, Map<Contract, Integer> contracts) throws ExecutionException, InterruptedException {
        Query queries = new Query(pred, mRules, mQueriesRe, mQueriesDe, parsedBytecodes, contracts, varMap, memBound, timeBound);
        if (testQueries){
            queries.ExecuteTestQueries(GUARD, contractID);
        }
        else {
            if (reentrancy){
                queries.CheckReentrancy(GUARD, contractID);
            }
            else {
                if (dependancy){
                    queries.CheckDependency(GUARD, contractID);
                }
                else {
                    queries.ExecuteQueries(GUARD, contractID);
                    //queries.ExecuteQueriesCheck(gasLimit, maxEvmValue, GUARD);
                }
            }
        }
    }
}
