package com.ethornium.analysis;

import com.ethornium.encoding.*;
import com.ethornium.encoding.enums.EVarArr;
import com.ethornium.encoding.enums.EVarArrS;
import com.ethornium.encoding.enums.EVarArrT;
import com.ethornium.encoding.enums.EVarInt;

import java.util.ArrayList;
import java.util.Map;
import java.util.Vector;

public class InitializingRules {
    private final ArrayList<EExpr> mRules;
    private final Pred pred;
    private final Vector<Integer> jumpDestinations;
    private final Integer maxPc;
    private final String bytecode;
    private final int contractID;
    private Map<Contract, Integer> contracts;
    private final VarMap varMap;
    private final boolean memBound;

    public InitializingRules(final ArrayList<EExpr> mRules, final Pred pred,
                             Vector<Integer> jumpDests,
                             Integer maxPc, final String bytecode,
                             final int contractID, Map<Contract, Integer> contracts,
                             VarMap varMap, boolean memBound) {
        this.mRules = mRules;
        this.jumpDestinations = jumpDests;
        this.maxPc = maxPc;
        this.bytecode = bytecode;
        this.contractID = contractID;
        this.contracts = contracts;
        this.pred = pred;
        this.varMap = varMap;
        this.memBound = memBound;
    }
    /*private void powerF(){
        //idp = exponent
        EExpr h, b;
        h = EEngine.eq(
                varMap.getIntExpr(EVarInt.idp),
                EEngine.num(0)
        );
        b = pred.powerF(varMap.getIntExpr(EVarInt.cp), varMap.getIntExpr(EVarInt.idp), EEngine.num(1));

        mRules.add(EEngine.implies(h, b));

        h = EEngine.and(
          EEngine.gt(

                  varMap.getIntExpr(EVarInt.idp),
                  EEngine.num(0)),

                pred.powerF(varMap.getIntExpr(EVarInt.cp),
                        EEngine.sub(varMap.getIntExpr(EVarInt.idp), EEngine.num(1)),
                        varMap.getIntExpr(EVarInt.gp))
        );

        b = pred.powerF(varMap.getIntExpr(EVarInt.cp), varMap.getIntExpr(EVarInt.idp), EEngine.mul(varMap.getIntExpr(EVarInt.cp), varMap.getIntExpr(EVarInt.gp)));
        mRules.add(EEngine.implies(h, b));
    }*/

    public void process() {
        //powerF();
        mRules.add(pred.gStatePred(Integer.toString(contractID),
                0,
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b),
                0));

        if (memBound) {
            mRules.add(pred.mStatePredMem(Integer.toString(contractID),
                    0,
                    EEngine.num(0),
                    varMap.getArrExpr(EVarArr.initstack),
                    varMap.getArrTExpr(EVarArrT.initstackT),
                    varMap.getArrSExpr(EVarArrS.initstackS),
                    EEngine.bool(false),
                    0, varMap.getMemZero()));
        }
        else{
            mRules.add(pred.mStatePred(Integer.toString(contractID),
                    0,
                    EEngine.num(0),
                    varMap.getArrExpr(EVarArr.initstack),
                    varMap.getArrTExpr(EVarArrT.initstackT),
                    varMap.getArrSExpr(EVarArrS.initstackS),
                    EEngine.bool(false),
                    0));

            mRules.add(pred.memPred(Integer.toString(contractID), 0, varMap.getIntExpr(EVarInt.pos), EEngine.num(0), EEngine.bool(false), EEngine.bool(false), 0));
        }
        /*mRules.add(pred.exenvPred(Integer.toString(contractID),
                varMap.getIntExpr(EVarInt.a),
                //EEngine.num("347890"), //hashcode of a current contract
                varMap.getIntExpr(EVarInt.i),
                0,
                varMap.getIntExpr(EVarInt.va),
                varMap.getIntExpr(EVarInt.size)));*/

        //initializing code predicate
        /*int i = 0;
        do {
            String op = bytecode.substring(i, i + 2);
            int opInt = Integer.decode("0x" + op);
            mRules.add(pred.codePred(Integer.toString(contractID),
                    new EExpr(i / 2),
                    new EExpr(opInt)));
            i = i + 2;

        } while (i < bytecode.length());*/


        //TODO: rework this, the code might get stuck
        //initializing calldata predicates
        /*mRules.add(pred.inputDataPred(Integer.toString(contractID),
                varMap.getIntExpr(EVarInt.val0),
                varMap.getIntExpr(EVarInt.val1),
                0
        ));*/

        /*mRules.add(pred.storPred(
                Integer.toString(contractID),
                0,
                //EEngine.num("347890"), //hashcode of a current contract
                varMap.getIntExpr(EVarInt.pos),
                EEngine.num(0),
                EEngine.bool(false),
                EEngine.bool(false),
                0));*/

        mRules.add(pred.storPred(
                Integer.toString(contractID),
                0,
                //EEngine.num("347890"), //hashcode of a current contract
                varMap.getIntExpr(EVarInt.pos),
                varMap.getIntExpr(EVarInt.val0),
                EEngine.bool(true),
                EEngine.bool(true),
                0));

    }
}
