package com.ethornium.encoding;

import com.ethornium.encoding.enums.EOperator;
import com.ethornium.encoding.enums.ESort;

public class EEngine {
    public static EExpr bool(boolean bul){
        return new EExpr(bul);
    }
    public static EExpr num(int num){
        return new EExpr(num);
    }
    public static EExpr num(String num){
        return new EExpr(num);
    }

    public static EExpr sub(EExpr v1, EExpr v2){
        return new EExpr(EOperator.sub, ESort.IntSort, v1, v2);
    }
    public static EExpr sub(EExpr v1, int num){
        EExpr e = num(num);
        return new EExpr(EOperator.sub, ESort.IntSort, v1, e);
    }
    public static EExpr add(EExpr v1, EExpr v2){
        return new EExpr(EOperator.add, ESort.IntSort, v1, v2);
    }
    public static EExpr add(EExpr v1, int num){
        EExpr e = num(num);
        return new EExpr(EOperator.add, ESort.IntSort, v1, e);
    }
    public static EExpr mul(EExpr v1, EExpr v2){
        return new EExpr(EOperator.mul, ESort.IntSort, v1, v2);
    }
    public static EExpr div(EExpr v1, EExpr v2){
        return new EExpr(EOperator.div, ESort.IntSort, v1, v2);
    }
    public static EExpr mod(EExpr v1, EExpr v2){
        return new EExpr(EOperator.mod, ESort.IntSort, v1, v2);
    }
    /*public static EExpr power(EExpr v1, EExpr v2){
        return new EExpr(EOperator.power, ESort.IntSort, v1, v2);
    }*/
    public static EExpr implies(EExpr v1, EExpr v2){
        return new EExpr(EOperator.implies, ESort.BoolSort, v1, v2);
    }
    public static EExpr selectInt(EExpr v1, EExpr v2){
        return new EExpr(EOperator.selectInt, ESort.ArraySortInt, v1, v2);
    }
    public static EExpr storeInt(EExpr v1, EExpr v2, EExpr v3){
        return new EExpr(EOperator.storeInt, ESort.ArraySortInt, v1, v2, v3);
    }
    public static EExpr selectT(EExpr v1, EExpr v2){
        return new EExpr(EOperator.selectT, ESort.ArraySortT, v1, v2);
    }
    public static EExpr selectS(EExpr v1, EExpr v2){
        return new EExpr(EOperator.selectS, ESort.ArraySortS, v1, v2);
    }
    public static EExpr storeT(EExpr v1, EExpr v2, EExpr v3){
        return new EExpr(EOperator.storeT, ESort.ArraySortT, v1, v2, v3);
    }
    public static EExpr storeS(EExpr v1, EExpr v2, EExpr v3){
        return new EExpr(EOperator.storeS, ESort.ArraySortS, v1, v2, v3);
    }
    public static EExpr and(EExpr... v){
        return new EExpr(EOperator.and, ESort.BoolSort, v);
    }
    public static EExpr or(EExpr... v){
        return new EExpr(EOperator.or, ESort.BoolSort, v);
    }
    public static EExpr not(EExpr v){
        return new EExpr(EOperator.not, ESort.BoolSort, v);
    }
    public static EExpr eq(EExpr v1, EExpr v2){
        return new EExpr(EOperator.eq, ESort.BoolSort, v1, v2);
    }
    /*public static EExpr checkValid(EExpr v1, EExpr v2){
        return new EExpr(EOperator.checkValid, ESort.BoolSort, v1, v2);
    } */
    public static EExpr ite(EExpr v1, EExpr v2, EExpr v3){
        return new EExpr(EOperator.ite, ESort.BoolSort, v1, v2, v3);
    }
    public static EExpr log256Floor(EExpr v){
        return new EExpr(EOperator.log256Floor, ESort.IntSort, v);
    }
    public static EExpr gt(EExpr v1, EExpr v2){
        return new EExpr(EOperator.gt, ESort.BoolSort, v1, v2);
    }
    public static EExpr le(EExpr v1, EExpr v2){
        return new EExpr(EOperator.le, ESort.BoolSort, v1, v2);
    }
    public static EExpr lt(EExpr v1, EExpr v2){
        return new EExpr(EOperator.lt, ESort.BoolSort, v1, v2);
    }
    public static EExpr ge(EExpr v1, EExpr v2){
        return new EExpr(EOperator.ge, ESort.BoolSort, v1, v2);
    }
    public static EExpr andBitwise(EExpr v1, EExpr v2){
        return new EExpr(EOperator.andBitwise, ESort.IntSort, v1, v2);
    }
    public static EExpr orBitwise(EExpr v1, EExpr v2){
        return new EExpr(EOperator.orBitwise, ESort.IntSort, v1, v2);
    }
    public static EExpr xorBitwise(EExpr v1, EExpr v2){
        return new EExpr(EOperator.xorBitwise, ESort.IntSort, v1, v2);
    }
    public static EExpr notBitwise(EExpr v){
        return new EExpr(EOperator.notBitwise, ESort.IntSort, v);
    }
    public static EExpr divCeiling(EExpr v1, EExpr v2){
        return new EExpr(EOperator.divCeiling, ESort.IntSort, v1, v2);
    }
    public static EExpr max(EExpr v1, EExpr v2){
        return new EExpr(EOperator.max, ESort.IntSort, v1, v2);
    }
    public static EExpr min(EExpr v1, EExpr v2){
        return new EExpr(EOperator.min, ESort.IntSort, v1, v2);
    }

    public static EExpr extract(EExpr v, int i, int j){
        EExpr e1 = num(i);
        EExpr e2 = num(j);
        return new EExpr(EOperator.extract, ESort.IntSort, v, e1, e2);
    }
    public static EExpr concat(EExpr v1, EExpr v2, int num){
        EExpr e = num(num);
        return new EExpr(EOperator.concat, ESort.IntSort, v1, v2, e);
    }
    public static EExpr cMem(EExpr v1, EExpr v2){
        return new EExpr(EOperator.cMem, ESort.IntSort, v1, v2);
    }
    public static EExpr M(EExpr i, EExpr o, EExpr s) { return new EExpr(EOperator.M, ESort.IntSort, i, o, s);}
}