package com.ethornium.encoding;

import com.ethornium.encoding.enums.*;

import java.util.HashMap;
import java.util.Map;

public class VarMap {
    private final int GUARD;
    private final int memSize;
    Map<EVarInt, Integer> varIntMap = new HashMap<>();
    Map<EVarArr, Integer> varArrMap = new HashMap<>();
    Map<Taint, Integer> taintMap = new HashMap<>();
    Map<Sym, Integer> symbolMap = new HashMap<>();
    Map<EVarArrT, Integer> varArrTaintMap = new HashMap<>();
    Map<EVarArrS, Integer> varArrSymbolMap = new HashMap<>();
    public EExpr getIntExpr(EVarInt var) {
        final Integer ind = varIntMap.get(var);
        return new EExpr(ESort.IntSort, ind);
    }
    public EExpr getArrExpr(EVarArr var) {
        final Integer ind = varArrMap.get(var);
        return new EExpr(ESort.ArraySortInt, ind);
    }
    public EExpr getTaintExpr(Taint var) {
        final Integer ind = taintMap.get(var);
        return new EExpr(ESort.Taint, ind);
    }
    public EExpr getSymbolicExpr(Sym var) {
        final Integer ind = symbolMap.get(var);
        return new EExpr(ESort.Symbolic, ind);
    }
    public EExpr getArrTExpr(EVarArrT var){
        final Integer ind = varArrTaintMap.get(var);
        return new EExpr(ESort.ArraySortT, ind);
    }
    public EExpr getArrSExpr(EVarArrS var){
        final Integer ind = varArrSymbolMap.get(var);
        return new EExpr(ESort.ArraySortS, ind);
    }
    public EExpr getVal(int i)  {
          return new EExpr(ESort.IntSort, GUARD + 3 * i + 1);
    }
    public EExpr getValT(int i) {
        return new EExpr(ESort.Taint, GUARD + 3 * i + 2);
    }
    public EExpr getValS(int i) {
        return new EExpr(ESort.Symbolic, GUARD + 3 * i + 3);
    }

    public EExpr[] getMemUp(final Map<Integer, EExpr> memUp){
        EExpr[] e = new EExpr[memSize];
        for (int i = 0; i < memSize; i++) {
            e[i] = memUp.get(i);
            if (e[i] == null) {
                e[i] = getVal(i);
            }
        }
        return e;
    }
    public EExpr[] getMem(){
        EExpr[] e = new EExpr[memSize];
        for (int i = 0; i < memSize; i++) {
            e[i] = getVal(i);
        }
        return e;
    }
    public EExpr[] getMemOver(){
        EExpr[] e = new EExpr[memSize];
        for (int i = 0; i < memSize; i++) {
            e[i] = getVal(i + memSize);
        }
        return e;
    }
    public EExpr[] getMemZero(){
        EExpr[] e = new EExpr[memSize];
        for (int i = 0; i < memSize; i++) {
            e[i] = new EExpr(0);
        }
        return e;
    }
    public VarMap(final int GUARD, final int memSize) {
        this.GUARD = GUARD;
        this.memSize = memSize;
        varIntMap.put(EVarInt.c, 0);
        varIntMap.put(EVarInt.pc, 1);
        varIntMap.put(EVarInt.g, 2);
        varIntMap.put(EVarInt.gas, 3);
        varIntMap.put(EVarInt.size, 4);
        varIntMap.put(EVarInt.aw, 5);
        varIntMap.put(EVarInt.a, 6);
        varIntMap.put(EVarInt.n, 7);
        varIntMap.put(EVarInt.b, 8);
        varIntMap.put(EVarInt.gc, 9);
        varIntMap.put(EVarInt.i, 10);
        varIntMap.put(EVarInt.x, 11);
        varIntMap.put(EVarInt.y, 12);
        varIntMap.put(EVarInt.ma, 13);
        varIntMap.put(EVarInt.awnew, 14);
        varIntMap.put(EVarInt.v, 15);
        varIntMap.put(EVarInt.sizeup, 16);
        varIntMap.put(EVarInt.sizelow, 17);
        varIntMap.put(EVarInt.posmem, 18);
        varIntMap.put(EVarInt.va, 19);
        varIntMap.put(EVarInt.cd, 20);
        varIntMap.put(EVarInt.posinp, 21);
        varIntMap.put(EVarInt.posStor, 22);
        varIntMap.put(EVarInt.vaStor, 23);
        varIntMap.put(EVarInt.jumpDest, 24);
        varIntMap.put(EVarInt.bcond, 25);
        varIntMap.put(EVarInt.cost, 26);
        varIntMap.put(EVarInt.varesult, 59);
        varIntMap.put(EVarInt.returndatasize, 60);
        varIntMap.put(EVarInt.aprime, 61);
        varIntMap.put(EVarInt.inputsize, 62);
        varIntMap.put(EVarInt.s, 63);
        varIntMap.put(EVarInt.pos, 64);
        varIntMap.put(EVarInt.poscom, 65);
        varIntMap.put(EVarInt.codesize, 66);
        varIntMap.put(EVarInt.com, 67);
        varIntMap.put(EVarInt.wpos, 68);
        varIntMap.put(EVarInt.wmem, 69);
        varIntMap.put(EVarInt.gaslimit, 89);
        varIntMap.put(EVarInt.stacklimit, 90);
        varIntMap.put(EVarInt.val0, 80);
        varIntMap.put(EVarInt.val1, 81);
        varIntMap.put(EVarInt.vnew, 82);
        varIntMap.put(EVarInt.newG, 83);
        varIntMap.put(EVarInt.newGas, 84);
        varIntMap.put(EVarInt.stackAddress, 86);

        varIntMap.put(EVarInt.gasTo, 30);
        varIntMap.put(EVarInt.to, 31);
        varIntMap.put(EVarInt.io, 32);
        varIntMap.put(EVarInt.is, 33);
        varIntMap.put(EVarInt.oo, 34);
        varIntMap.put(EVarInt.os, 35);
        varIntMap.put(EVarInt.cUpfront, 36);
        varIntMap.put(EVarInt.cex, 37);
        varIntMap.put(EVarInt.cGascap, 38);
        varIntMap.put(EVarInt.CCall, 39);
        varIntMap.put(EVarInt.ccase, 40);
        varIntMap.put(EVarInt.idp, 41);
        varIntMap.put(EVarInt.cp, 42);
        varIntMap.put(EVarInt.k, 43);
        varIntMap.put(EVarInt.gp, 44);
        varIntMap.put(EVarInt.code, 45);
        varIntMap.put(EVarInt.id, 46);

        varIntMap.put(EVarInt.z, 47);

        varArrMap.put(EVarArr.newStack, 85);
        varArrMap.put(EVarArr.stack, 70);
        varArrMap.put(EVarArr.initstack, 91);

        varArrTaintMap.put(EVarArrT.newStackT, 92);
        varArrTaintMap.put(EVarArrT.stackT, 93);
        varArrTaintMap.put(EVarArrT.initstackT, 94);
        varArrSymbolMap.put(EVarArrS.newStackS, 95);
        varArrSymbolMap.put(EVarArrS.stackS, 96);
        varArrSymbolMap.put(EVarArrS.initstackS, 97);

        taintMap.put(Taint.taint, 100);
        taintMap.put(Taint.xt, 101);
        taintMap.put(Taint.yt, 102);
        taintMap.put(Taint.zt, 103);
        taintMap.put(Taint.rezt, 104);
        taintMap.put(Taint.taintM, 105);
        taintMap.put(Taint.taintS, 106);

        symbolMap.put(Sym.sym, 120);
        symbolMap.put(Sym.xs, 121);
        symbolMap.put(Sym.ys, 122);
        symbolMap.put(Sym.zs, 123);
        symbolMap.put(Sym.rezs, 124);
        symbolMap.put(Sym.symM, 125);
        symbolMap.put(Sym.symS, 126);

    }
}
