package com.ethornium.util;
import java.io.*;
import java.math.BigInteger;
import java.net.*;
import java.util.*;
import java.util.logging.Logger;

import com.ethornium.encoding.Contract;
import org.json.*;
import org.json.simple.parser.ParseException;
import org.json.simple.JSONObject;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import org.json.simple.parser.JSONParser;

public class GetBytecode {
    private static final String apiKey="RWM6D4DV2C2PWY3CFWV3F81A38X1WHPYBY";

    public static org.json.JSONObject getData(String targetURL) throws IOException{
        StringBuilder json = new StringBuilder();
        URL url = new URL(targetURL);
        HttpURLConnection conn = (HttpURLConnection) url.openConnection();
        conn.setRequestMethod("GET");
        BufferedReader rd = new BufferedReader(new InputStreamReader(conn.getInputStream()));
        String line;
        while ((line = rd.readLine()) != null) {
            json.append(line);
        }
        rd.close();
        org.json.JSONObject obj = new org.json.JSONObject(json.toString());
        return obj;
    }

    public static String fromHexToInt(String hex){
        BigInteger bi = new BigInteger(hex.replace("0x",""), 16);
        String rez = bi.toString();
        return rez;
    }

    public static String getBytecodeFromEtherscan(String address) throws IOException {
        String targetURL="https://api.etherscan.io/api?module=proxy&action=eth_getCode&address="
                + address + "&tag=latest&apikey=" + apiKey;
        org.json.JSONObject obj = getData(targetURL);
        return obj.get("result").toString();
    }

    public static String getBlockNumber() throws IOException {
        String targetURL="https://api.etherscan.io/api?module=proxy&action=eth_blockNumber&apikey=YourApiKeyToken&apikey=" + apiKey;
        org.json.JSONObject obj = getData(targetURL);
        return obj.get("result").toString();
    }

    public static Map<String, Object> jsonToMap(org.json.JSONObject json) throws JSONException {
        Map<String, Object> retMap = new HashMap();
        if(json != org.json.JSONObject.NULL) {
            retMap = toMap(json);
        }
        return retMap;
    }

    public static Map<String, Object> toMap(org.json.JSONObject object) throws JSONException {
        Map<String, Object> map = new HashMap();
        Iterator<String> keysItr = object.keys();
        while(keysItr.hasNext()) {
            String key = keysItr.next();
            Object value = object.get(key);
            if(value instanceof JSONArray) {
                value = toList((JSONArray) value);
            }
            else if(value instanceof JSONObject) {
                value = toMap((org.json.JSONObject) value);
            }
            map.put(key, value);
        }
        return map;
    }

    public static List<Object> toList(JSONArray array) throws JSONException {
        List<Object> list = new ArrayList();
        for(int i = 0; i < array.length(); i++) {
            Object value = array.get(i);
            if(value instanceof JSONArray) {
                value = toList((JSONArray) value);
            }
            else if(value instanceof JSONObject) {
                value = toMap((org.json.JSONObject) value);
            }
            list.add(value);
        }
        return list;
    }

    public static Contract getContractData(String address, Contract c) throws IOException{
        c.assignAddress(fromHexToInt(address));
        String targetURL="https://api.etherscan.io/api?module=account&action=balance&address="
                + address + "&tag=latest&apikey=" + apiKey;
        org.json.JSONObject obj = getData(targetURL);
        String balance = obj.get("result").toString();
        c.assignBalance(balance);

        targetURL="https://api.etherscan.io/api?module=contract&action=getabi&address="
                + address + "&tag=latest&apikey=" + apiKey;
        obj = getData(targetURL);
        String abi = obj.get("result").toString();
        c.assignAbi(abi);

        return c;
    }

    public static void getTopContracts() throws IOException {
        PrintWriter writer = new PrintWriter("/Users/ilya/Desktop/EtherScanTop.txt", "UTF-8");
        for (int i = 1; i <= 400; i++) {
            String url = "https://etherscan.io/accounts/" + Integer.toString(i);
            try {
                Document document = Jsoup.connect(url).get();
                ArrayList<String> topContracts = new ArrayList<>();
                Elements links = document.select("a");
                for (Element link : links) {
                    String addr = link.attr("href");
                    if (addr != null) {
                        if (addr.contains("address/0x")) {
                            String[] parts = addr.split("/address/");
                            String eAddr = parts[1];
                            if (!eAddr.isEmpty()) {
                                if (eAddr.contains("#comments")) {
                                    String[] partsCom = eAddr.split("#comments");
                                    String eAddrCom = partsCom[0];
                                    topContracts.add(eAddrCom);
                                    writer.println(eAddrCom);
                                } else {
                                    topContracts.add(eAddr);
                                    writer.println(eAddr);
                                }
                            }
                        }
                    }
                }
            }
            catch (Exception e){
                int j = 0;
                j  = j + 1;
            }
        }

        writer.close();
    }

    public static void makeDatabase() throws IOException {
        BufferedReader reader = new BufferedReader(new FileReader("/Users/ilya/Desktop/EtherScanTop_0405.txt"));
        int lines = 0;
        String line;
        while ((line = reader.readLine()) != null) {
            String address = line;
            File f = new File("/Users/ilya/Desktop/database/"+address+".json");
            if (f.exists()) continue;
            String[] commandBalance = {"curl","localhost:8545", "-H", "Content-Type: application/json",
                    "-X", "POST" ,"--data", "{\"jsonrpc\":\"2.0\",\"method\":\"eth_getBalance\",\"params\":[\"" + address +
                    "\", \"latest\"],\"id\":1}"};
            String balance = fromHexToInt(execCommand(commandBalance));
            String[] commandCode = {"curl","localhost:8545", "-H", "Content-Type: application/json",
                    "-X", "POST" ,"--data", "{\"jsonrpc\":\"2.0\",\"method\":\"eth_getCode\",\"params\":[\"" + address +
                    "\", \"latest\"],\"id\":1}"};
            //String abi = null;
            String bytecode = execCommand(commandCode);

            org.json.JSONObject jsonO = new org.json.JSONObject();
            jsonO.put("address", address);
            jsonO.put("balance", balance);
            jsonO.put("bytecode", bytecode);

            try (FileWriter file = new FileWriter("/Users/ilya/Desktop/database/"+address+".json")) {
                file.write(jsonO.toString(2));
                file.flush();

            } catch (IOException e) {
                e.printStackTrace();
            }
        }
        reader.close();
    }

    public static String execCommand(String[] command){
        ProcessBuilder process = new ProcessBuilder(command);
        Process p;
        String result;
        org.json.JSONObject objj;
        try
        {
            p = process.start();
            BufferedReader reader =  new BufferedReader(new InputStreamReader(p.getInputStream()));
            StringBuilder builder = new StringBuilder();
            String line;
            while ( (line = reader.readLine()) != null) {
                builder.append(line);
                builder.append(System.getProperty("line.separator"));
            }
            result = builder.toString();
            objj = new org.json.JSONObject(result);
            return objj.get("result").toString();
        }
        catch (IOException e)
        {   System.out.print("error");
            e.printStackTrace();
        }
        return null;
    }

    public static Block getBlockInfoFromJsonFile() throws IOException, ParseException {
        JSONParser parser = new JSONParser();

        JSONObject jsonObject = (JSONObject) parser.parse(new FileReader("block.json"));

            //JSONObject jsonObject = (JSONObject) obj;

        String totalDifficulty = (String) jsonObject.get("totalDifficulty");
        String nonce = (String) jsonObject.get("nonce");
        String difficulty = (String) jsonObject.get("difficulty");

        String gasLimit = (String) jsonObject.get("gasLimit");
        String number = (String) jsonObject.get("number");
        String size = (String) jsonObject.get("size");

        String hash = (String) jsonObject.get("hash");
        String timestamp = (String) jsonObject.get("timestamp");
        String receiptsRoot = (String) jsonObject.get("receiptsRoot");
        String gasPrice = (String) jsonObject.get("gasPrice");

        return new Block(totalDifficulty, nonce, difficulty, gasLimit, number, size, hash, timestamp, receiptsRoot, gasPrice);
    }

    public static Contract getContractInfoFromJsonFile(String path) throws IOException, ParseException {
        JSONParser parser = new JSONParser();

        Object obj = parser.parse(new FileReader(path));

        JSONObject jsonObject = (JSONObject) obj;

        String address = fromHexToInt((String) jsonObject.get("address"));
        String balance = (String) jsonObject.get("balance");
        String bytecode = (String) jsonObject.get("bytecode");

        Contract c = new Contract((String) jsonObject.get("address"), 0, bytecode);
        c.assignAddress(address);
        c.assignBalance(balance);
        return c;
    }

    public static Block getBlockInfoFromEtherscan(String address) throws IOException {
        String blockNumber = getBlockNumber();
        String targetURL = "https://api.etherscan.io/api?module=proxy&action=eth_getBlockByNumber&tag="
                + blockNumber + "&boolean=true&apikey=" + apiKey;
        org.json.JSONObject obj = getData(targetURL);
        Map<String, Object> result = jsonToMap((org.json.JSONObject) obj.get("result"));
        String totalDifficulty = fromHexToInt((String) result.get("totalDifficulty"));
        String nonce = fromHexToInt((String) result.get("nonce"));
        String difficulty = fromHexToInt((String) result.get("difficulty"));
        String gasLimit = fromHexToInt((String) result.get("gasLimit"));
        String number = fromHexToInt((String) result.get("number"));
        String size = fromHexToInt((String) result.get("size"));
        String hash = fromHexToInt((String) result.get("hash"));
        String timestamp = fromHexToInt((String) result.get("timestamp"));
        String receiptsRoot = fromHexToInt((String) result.get("receiptsRoot"));

        targetURL="https://api.etherscan.io/api?module=proxy&action=eth_gasPrice&apikey=" + apiKey;
        obj = getData(targetURL);
        String gasPrice = fromHexToInt(obj.get("result").toString());

        org.json.JSONObject jsonO = new org.json.JSONObject();
        jsonO.put("totalDifficulty", totalDifficulty);
        jsonO.put("nonce", nonce);
        jsonO.put("difficulty", difficulty);
        jsonO.put("gasLimit", gasLimit);
        jsonO.put("number", number);
        jsonO.put("size", size);
        jsonO.put("hash", hash);
        jsonO.put("timestamp", timestamp);
        jsonO.put("receiptsRoot", receiptsRoot);
        jsonO.put("gasPrice", gasPrice);

        try (FileWriter file = new FileWriter("/Users/ilya/Desktop/block.json")) {
            file.write(jsonO.toString(2));
            file.flush();

        } catch (IOException e) {
            e.printStackTrace();
        }

        return new Block(totalDifficulty, nonce, difficulty, gasLimit, number, size, hash, timestamp, receiptsRoot, gasPrice);
    }
}
