package com.ethornium.z3;

import com.microsoft.z3.*;
import org.apache.logging.log4j.LogManager;

import java.math.BigInteger;

public class Engine {
    private final Context ctx;
    private final BigInteger two = new BigInteger("2");
    private static final org.apache.logging.log4j.Logger LOGGER = LogManager.getLogger(Engine.class);

    private IntExpr modConst(){
        return num("115792089237316195423570985008687907853269984665640564039457584007913129639936");
    }

    public Engine(final Context ctx){
        this.ctx = ctx;
    }

    public BoolExpr tru(){
        return ctx.mkTrue();
    }

    public BoolExpr fal(){
        return ctx.mkFalse();
    }

    public BoolExpr implies(BoolExpr head, BoolExpr body){
        return ctx.mkImplies(head, body);
    }

    public BoolExpr gt(IntExpr v1, IntExpr v2){
        return ctx.mkGt(v1, v2);
    }

    public BoolExpr lt(IntExpr v1, IntExpr v2){
        return ctx.mkLt(v1, v2);
    }

    public BoolExpr le(IntExpr v1, IntExpr v2){
        return ctx.mkLe(v1, v2);
    }

    public BoolExpr ge(IntExpr v1, IntExpr v2){
        return ctx.mkGe(v1, v2);
    }

    public BoolExpr not (BoolExpr b) { return ctx.mkNot(b); }

    public BoolExpr eq (Expr b1, Expr b2) { return  ctx.mkEq(b1, b2); }

    public BoolExpr and(BoolExpr... b){
        return ctx.mkAnd(b);
    }

    public BoolExpr or(BoolExpr... b){
        return ctx.mkOr(b);
    }

    //public ArrayExpr store (ArrayExpr a, IntExpr i, IntExpr j){
    //    return ctx.mkStore(a, i, j);
    //}

    public ArrayExpr store (ArrayExpr a, IntExpr i, Expr j){
        return ctx.mkStore(a, i, j);
    }

    public IntExpr mod (IntExpr v1, IntExpr v2) { return ctx.mkMod(ctx.mkMod(v1, v2), modConst()); }

    public IntExpr selectInt (ArrayExpr a, IntExpr i) { return (IntExpr) ctx.mkSelect(a, i); }
    public BoolExpr selectBool (ArrayExpr a, IntExpr i) { return (BoolExpr) ctx.mkSelect(a, i); }

    //public IntExpr select (ArrayExpr a, ArithExpr i) { return (IntExpr) ctx.mkSelect(a, i); }

    public IntExpr add(IntExpr a, Integer b) {
        return ctx.mkMod((IntExpr) ctx.mkAdd(a, ctx.mkInt(b)), modConst());
    }

    public IntExpr add(IntExpr a, IntExpr b) {
        return  ctx.mkMod((IntExpr) ctx.mkAdd(a, b), modConst());
    }

    public IntExpr mulConst(IntExpr a, int b) {
        IntExpr rez = num (0);
        /*for (int i = 0; i < b ; i++){
            rez = add(rez, a);
        }*/
        return rez;
    }

    public IntExpr mul(IntExpr a, IntExpr b) {
        return ctx.mkMod((IntExpr) ctx.mkMul(a, b), modConst());
    }

    public IntExpr div(IntExpr a, IntExpr b) {
        return ctx.mkMod((IntExpr) ctx.mkDiv(a, b), modConst());
    }

    public Expr ite (BoolExpr b, Expr x, Expr y){
        return ctx.mkITE(b, x, y);
    }

    //public ArithExpr power (IntExpr a, IntExpr b) { return ctx.mkPower(a, b);}

    public IntExpr sub(IntExpr a, int b) {
        return ctx.mkMod((IntExpr) ctx.mkSub(a, ctx.mkInt(b)), modConst());
    }

    public IntExpr sub(IntExpr a, IntExpr b) {
        return ctx.mkMod((IntExpr) ctx.mkSub(a, b), modConst());
    }

    public IntExpr num(Integer a) {
        return ctx.mkInt(a);
    }

    public IntExpr num(long a) {
        return ctx.mkInt(a);
    }

    public IntExpr num(String a) {
        return ctx.mkInt(a);
    }

    public IntExpr max(IntExpr a, IntExpr b) {
        return (IntExpr) ctx.mkITE(ctx.mkGe(a, b), a, b);
    }

    public IntExpr min(IntExpr a, IntExpr b) {
        return (IntExpr) ctx.mkITE(ctx.mkGe(a, b), b, a);
    }

    public IntExpr divCeiling(IntExpr a, IntExpr b) {
        return (IntExpr) this.ite(
                this.eq(
                        this.mod(a, b),
                        num(0)
                ),
                this.div(a, b),
                this.add(this.num(1), this.div(a, b)));
    }

    //returns the floor of log(a), assume that a is integer and greater than 0
    /*public IntExpr log256Floor(IntExpr a) {
        return (IntExpr) ctx.mkITE(
                ctx.mkLt(a, num(256)), num(0),
                ctx.mkITE(ctx.mkLt(a, num(65536)), num(1),
                ctx.mkITE(ctx.mkLt(a, num(16777216)), num(2),
                        ctx.mkITE(ctx.mkLt(a, num("4294967296")), num(3),
                        ctx.mkITE(ctx.mkLt(a, num("1099511627776")), num(4),
                                ctx.mkITE(ctx.mkLt(a, num("281474976710656")),
                                num(5),
                                        num(6)))))));
    }*/

    public IntExpr log256Floor(IntExpr a) {
        return (IntExpr) ite(
                lt(a, num("340282366920938463463374607431768211456")),
                    ite(
                            lt(a, num("18446744073709551616")),
                            ite(
                                    lt(a, num("4294967296")),
                                    ite(
                                            lt(a, num(65536)),
                                            ite(
                                                    lt(a, num(256)),
                                                    num(0),
                                                    num(1)
                                            ),
                                            ite(
                                                    lt(a, num("16777216")),
                                                    num(2),
                                                    num(3)
                                            )
                                    ),
                                    ite(
                                            lt(a,  num("281474976710656")),
                                            ite(
                                                    lt(a, num("1099511627776")),
                                                    num(4),
                                                    num(5)
                                            ),
                                            ite(
                                                    lt(a, num("72057594037927936")),
                                                    num(6),
                                                    num(7)
                                            )
                                    )
                            ),
                            ite(
                                    lt(a, num("79228162514264337593543950336")),
                                    ite(
                                            lt(a, num("1208925819614629174706176")),
                                            ite(
                                                    lt(a, num("4722366482869645213696")),
                                                    num(8),
                                                    num(9)
                                            ),
                                            ite(
                                                    lt(a, num("309485009821345068724781056")),
                                                    num(10),
                                                    num(11)
                                            )
                                    ),
                                    ite(
                                            lt(a, num("5192296858534827628530496329220096")),
                                            ite(
                                                    lt(a, num("20282409603651670423947251286016")),
                                                    num(12),
                                                    num(13)
                                            ),
                                            ite(
                                                    lt(a, num("1329227995784915872903807060280344576")),
                                                    num(14),
                                                    num(15)
                                            )
                                    )
                            )
                    ),
                ite(
                        lt(a, num("6277101735386680763835789423207666416102355444464034512896")),
                        ite(
                               lt(a, num("1461501637330902918203684832716283019655932542976")),
                                       ite(
                                               lt(a, num("22300745198530623141535718272648361505980416")),
                                               ite(
                                                       lt(a, num("87112285931760246646623899502532662132736")),
                                                       num(16),
                                                       num(17)
                                               ),
                                               ite(
                                                       lt(a, num("5708990770823839524233143877797980545530986496")),
                                                       num(18),
                                                       num(19)
                                               )
                                       ),
                                       ite(
                                               lt(a, num("95780971304118053647396689196894323976171195136475136")),
                                               ite(
                                                       lt(a, num("374144419156711147060143317175368453031918731001856")),
                                                       num(20),
                                                       num(21)
                                               ),
                                               ite(
                                                       lt(a, num("24519928653854221733733552434404946937899825954937634816")),
                                                       num(22),
                                                       num(23)
                                               )

                                       )
                        ),
                        ite(
                                lt(a, num("26959946667150639794667015087019630673637144422540572481103610249216")),
                                ite(
                                        lt(a, num("411376139330301510538742295639337626245683966408394965837152256")),
                                        ite(
                                                lt(a, num("1606938044258990275541962092341162602522202993782792835301376")),
                                                num(24),
                                                num(25)
                                        ),
                                        ite(
                                                lt(a, num("105312291668557186697918027683670432318895095400549111254310977536")),
                                                num(26),
                                                num(27)
                                        )
                                ),
                                ite(
                                        lt(a, num(
                                                "1766847064778384329583297500742918515827483896875618958121606201292619776")),
                                        ite(
                                                lt(a, num("6901746346790563787434755862277025452451108972170386555162524223799296")),
                                                num(28),
                                                num(29)
                                        ),
                                        ite(
                                                lt(a, num("452312848583266388373324160190187140051835877600158453279131187530910662656")),
                                                num(30),
                                                num(31)
                                        )
                                )
                        )
                )
        );
    }

//    public BoolExpr validF(IntExpr g, IntExpr gas, IntExpr sizeup, IntExpr sizelow) {
//            return (BoolExpr) ctx.mkITE(ctx.mkLt(g, num(0)), ctx.mkBool(false), ctx.mkITE(ctx.mkGt(g, gasLimit),
//                    ctx.mkBool(false), ctx.mkITE(ctx.mkLt(gas, num(0)), ctx.mkBool(false), ctx.mkITE(ctx.mkGt(gas, g),
//                            ctx.mkBool(false), ctx.mkITE(ctx.mkLt(sizelow, num(0)), ctx.mkBool(false),
//                                    ctx.mkITE(ctx.mkGt(sizeup, num(1024)), ctx.mkBool(false), ctx.mkBool(true)))))));
//    }

    public IntExpr notBitwise(IntExpr a) {
        /*BitVecExpr bvA = ctx.mkInt2BV(256, a);
        BitVecExpr not = ctx.mkBVNot(bvA);
        return ctx.mkBV2Int(not, false);*/

        IntExpr helper = ctx.mkInt(0);
        for (int n = 0; n < 256; n++) {
            IntExpr twoton = num(Math.round(Math.pow(2, n)));
            helper = (IntExpr) ctx.mkAdd(helper, ctx.mkMul(twoton, ctx.mkMod((IntExpr) ctx.mkAdd(ctx.mkMod((IntExpr)
                    ctx.mkDiv(a, twoton), num(2)), num(1)), num(2))));
        }
        return helper;

    }

    public IntExpr andBitwise(IntExpr a, IntExpr b) {
        /*BitVecExpr bvA = ctx.mkInt2BV(256, a);
        BitVecExpr bvB = ctx.mkInt2BV(256, b);
        BitVecExpr rez = ctx.mkBVAND(bvA, bvB);
        return ctx.mkBV2Int(rez, false);*/

        IntExpr helper = ctx.mkInt(0);
        for (int n = 0; n < 256; n++) {
            IntExpr twoton = num(Math.round(Math.pow(2, n)));
            helper = (IntExpr)
                    ctx.mkAdd(
                            helper,
                            ctx.mkMul(
                                    twoton,
                                    ctx.mkMul
                                            (ctx.mkMod((IntExpr) ctx.mkDiv(a, twoton), num(2)),
                                             ctx.mkMod((IntExpr) ctx.mkDiv(b, twoton), num(2))
                                            )
                            )
                    );
        }
        return helper;
    }

    public IntExpr orBitwise(IntExpr a, IntExpr b) {
        /*BitVecExpr bvA = ctx.mkInt2BV(256, a);
        BitVecExpr bvB = ctx.mkInt2BV(256, b);
        BitVecExpr rez = ctx.mkBVOR(bvA, bvB);
        return ctx.mkBV2Int(rez, false);*/

        IntExpr helper = ctx.mkInt(0);
        for (int n = 0; n < 256; n++) {
            IntExpr twoton = num(Math.round(Math.pow(2, n)));
            helper = (IntExpr) ctx.mkAdd(helper, ctx.mkMul(twoton, ctx.mkMod((IntExpr)
                    ctx.mkAdd(ctx.mkAdd(ctx.mkMod((IntExpr) ctx.mkDiv(a, twoton), num(2)), ctx.mkMod((IntExpr)
                            ctx.mkDiv(b, twoton), num(2))), ctx.mkMul(ctx.mkMod((IntExpr)
                            ctx.mkDiv(a, twoton), num(2)), ctx.mkMod((IntExpr)
                            ctx.mkDiv(b, twoton), num(2)))), num(2))));
        }
        return helper;
    }

    public IntExpr xorBitwise(IntExpr a, IntExpr b) {
        /*BitVecExpr bvA = ctx.mkInt2BV(256, a);
        BitVecExpr bvB = ctx.mkInt2BV(256, b);
        BitVecExpr rez = ctx.mkBVXOR(bvA, bvB);
        return ctx.mkBV2Int(rez, false);*/
        IntExpr helper = ctx.mkInt(0);
        for (int n = 0; n < 256; n++) {
            IntExpr twoton = num(Math.round(Math.pow(2, n)));
            helper = (IntExpr) ctx.mkAdd(helper, ctx.mkMul(twoton, ctx.mkMod((IntExpr)
                    ctx.mkAdd(ctx.mkMod((IntExpr) ctx.mkDiv(a, twoton), num(2)), ctx.mkMod((IntExpr)
                            ctx.mkDiv(b, twoton), num(2))), num(2))));
        }
        return helper;
    }

    public IntExpr extract(final IntExpr v, final int left, final int right){
        if (right < left || right < 0 || right > 31 || left < 0) {
            LOGGER.error("In extractor up address is less than down!");
            throw new IndexOutOfBoundsException();
        }
        IntExpr div = this.num(two.pow((31 - right) * 8).toString());
        IntExpr mod = this.num(two.pow((right - left) * 8).toString());
        IntExpr res = this.mod(div(v, div), mod);
        return res;
    }

    public IntExpr concat(final IntExpr v1, final IntExpr v2, final int l){
        IntExpr mul = this.num(two.pow(l * 8).toString());
        return this.add(this.mul(v1, mul), v2);
    }

    public IntExpr M(IntExpr i, IntExpr o, IntExpr s){
        return (IntExpr)
                this.ite(
                            this.eq(s, this.num(0)),
                            i,
                            this.max(i,
                                    this.divCeiling(
                                        this.add(o, s),
                                        this.num(32)
                                    )
                            )
                );
    }

    public IntExpr cMem(IntExpr v1, IntExpr v2){
        return this.add(
                this.mul(
                        this.num(3),
                        this.sub(v2, v1)
                ),
                this.sub(
                        this.div(
                                this.mul(v2, v2),
                                this.num(512)
                        ),
                        this.div(
                                this.mul(v1, v1),
                                this.num(512)
                        )
                )
        );
    }

    /*public BoolExpr checkValid(IntExpr sizeUP, IntExpr sizeLow){
        return (BoolExpr) ite(
                                                ge(sizeLow, num(0)),
                                                ite(
                                                        le(sizeUP, num(1024)),
                                                        tru(),
                                                        fal()
                                                ),
                                                fal()

        );
    }*/

//    public BoolExpr checkValidOld(IntExpr g, IntExpr gas, IntExpr sizeUP, IntExpr sizeLow){
//        return (BoolExpr) ite(
//                ge(g, num(0)),
//                ite(
//                        le(g, gasLimit),
//                        ite(ge(g, num(0)),
//                                ite(le(gas, g),
//                                        ite(
//                                           ge(sizeLow, num(0)),
//                                                ite(
//                                                        le(sizeUP, num(1024)),
//                                                        tru(),
//                                                        fal()
//                                                ),
//                                                fal()
//                                        ),
//                                        fal()
//
//                                ),
//                                fal()
//
//                        ),
//                        fal()
//                ),
//                fal()
//        );
//    }

    //public IntExpr getMaxEvmValue(){
    //    return ctx.mkInt(maxEvmValue);
    //}
}
