package com.ethornium.analysis;

import com.ethornium.encoding.*;
import com.ethornium.encoding.enums.*;
import com.ethornium.util.Block;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.math.BigInteger;
import java.util.*;


public class HornClauseGenerator {
    private static final Logger LOGGER = LogManager.getLogger(HornClauseGenerator.class);
    private final ArrayList<EExpr> mRules;
    private final ArrayList<EExpr> mQueriesRe;
    private final ArrayList<EExpr> mQueriesDe;
    private final Pred pred;
    private final Map<Integer, InternalRepresentation> internalRepresentation;
    private final Vector<Integer> jumpDests;
    private final Integer maxPc;
    private final String contractID;
    private final Map<Contract, Integer> contracts;
    private InternalRepresentation bytecode;
    private InternalRepresentation nextBytecode;
    private Integer pc;
    private Integer next;
    private final VarMap varMap;
    private EExpr x, y , z, rez, xt, yt, zt, rezt, xs, ys, zs, rezs;
    private final EExpr[] standardMem;
    private final EExpr[] ovappxMem;
    private final boolean debug;
    private final int memSize;
    private final boolean memBound;
    private final Integer cd;
    private final Contract contr;
    private final Block block;
    private final boolean reentrCheck;
    private final boolean depCheck;
    private final boolean addQueries;

    public HornClauseGenerator(final ArrayList<EExpr> mRules, final ArrayList<EExpr> mQueriesRe, final ArrayList<EExpr> mQueriesDe,  final Pred pred,
                               Map<Integer, InternalRepresentation> internalRepresentation,
                               Vector<Integer> jumpDests,
                               Integer maxPc,
                               final int contractID,
                               Map<Contract, Integer> contracts,
                               final VarMap varMap,
                               boolean  debug, int memSize, boolean memBound, Integer cd, Contract contr, Block block, boolean reentrCheck, boolean depCheck, boolean addQueries) {
        this.mRules = mRules;
        this.mQueriesRe = mQueriesRe;
        this.mQueriesDe = mQueriesDe;
        this.pred = pred;
        this.internalRepresentation = internalRepresentation;
        this.jumpDests = jumpDests;
        this.maxPc = maxPc;
        this.contractID = Integer.toString(contractID);
        this.contracts = contracts;
        this.bytecode = null;
        this.nextBytecode = null;
        this.pc = 0;
        this.next = null;
        this.varMap = varMap;
        this.standardMem = varMap.getMem();
        this.ovappxMem = varMap.getMemOver();
        this.debug = debug;
        this.memSize = memSize;
        this.memBound = memBound;
        this.cd = cd;
        this.contr = contr;
        this.block = block;
        this.reentrCheck = reentrCheck;
        this.depCheck = depCheck;
        this.addQueries = addQueries;
    }

    public void GenerateHornClauses() {
        Iterator it = internalRepresentation.entrySet().iterator();
        Iterator it2 = internalRepresentation.entrySet().iterator();

        while(it.hasNext()){
            Map.Entry<Integer, InternalRepresentation> pair = (Map.Entry<Integer, InternalRepresentation>) it.next();
            pc = pair.getKey();
            bytecode = pair.getValue();
            if (pc == 0) {
                it2.next();
            }
            Map.Entry<Integer, InternalRepresentation> nextPair;
            if (it.hasNext()){
                nextPair = (Map.Entry<Integer, InternalRepresentation>) it2.next();
                nextBytecode = nextPair.getValue();
                next = nextPair.getKey();
            }
            else{
                if (debug) {
                    LOGGER.debug("No next when expected, normal on tests");
                }
                next = pc + 1;
            }
            caller();
        }
    }

    EExpr getXArg(){
        if (bytecode.getComputedArguments().size()>0) {
            if (bytecode.getComputedArguments().get(0) != null) {
                return
                        EEngine.eq(
                                varMap.getIntExpr(EVarInt.x),
                                EEngine.num(bytecode.getComputedArguments().get(0))
                        );
            }
            else{
                return EEngine.eq(
                        varMap.getIntExpr(EVarInt.x),
                        EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
                );
            }
        }
        else{
            return EEngine.eq(
                    varMap.getIntExpr(EVarInt.x),
                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
            );
        }
    }

    EExpr getYArg(){
        if (bytecode.getComputedArguments().size()>0) {
            if (bytecode.getComputedArguments().get(1) != null) {
                return
                        EEngine.eq(
                                varMap.getIntExpr(EVarInt.y),
                                EEngine.num(bytecode.getComputedArguments().get(1))
                        );
            }
            else{
                return EEngine.eq(
                        varMap.getIntExpr(EVarInt.y),
                        EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2))
                );
            }
        }
        else{
            return EEngine.eq(
                    varMap.getIntExpr(EVarInt.y),
                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2))
            );
        }
    }

    EExpr getZArg(){
        if (bytecode.getComputedArguments().size()>0) {
            if (bytecode.getComputedArguments().get(2) != null) {
                return
                        EEngine.eq(
                                varMap.getIntExpr(EVarInt.z),
                                EEngine.num(bytecode.getComputedArguments().get(2))
                        );
            }
            else{
                return EEngine.eq(
                        varMap.getIntExpr(EVarInt.z),
                        EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 3))
                );
            }
        }
        else{
            return EEngine.eq(
                    varMap.getIntExpr(EVarInt.z),
                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3))
            );
        }
    }

    EExpr getXArgT(){
            return EEngine.eq(
                    varMap.getTaintExpr(Taint.xt),
                    EEngine.selectT(varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
            );
    }

    EExpr getYArgT(){
        return EEngine.eq(
                varMap.getTaintExpr(Taint.yt),
                EEngine.selectT(varMap.getArrTExpr(EVarArrT.stackT),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 2))
        );
    }

    EExpr getZArgT(){
        return EEngine.eq(
                varMap.getTaintExpr(Taint.zt),
                EEngine.selectT(varMap.getArrTExpr(EVarArrT.stackT),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 3))
        );
    }

    EExpr getXArgS(){
        return EEngine.eq(
                varMap.getSymbolicExpr(Sym.xs),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );
    }

    EExpr getYArgS(){
        return EEngine.eq(
                varMap.getSymbolicExpr(Sym.ys),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 2))
        );
    }

    EExpr getZArgS(){
        return EEngine.eq(
                varMap.getSymbolicExpr(Sym.zs),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 3))
        );
    }


    EExpr getREZTJoinXY(){
        return EEngine.eq(
                varMap.getTaintExpr(Taint.rezt),
                EEngine.or(varMap.getTaintExpr(Taint.xt), varMap.getTaintExpr(Taint.yt))
        );
    }

    EExpr getREZTJoinXYZ(){
        return EEngine.eq(
                varMap.getTaintExpr(Taint.rezt),
                EEngine.or(varMap.getTaintExpr(Taint.xt), varMap.getTaintExpr(Taint.yt), varMap.getTaintExpr(Taint.zt))
        );
    }

    EExpr getREZSJoinXY(){
        return EEngine.eq(
                varMap.getSymbolicExpr(Sym.rezs),
                EEngine.or(varMap.getSymbolicExpr(Sym.xs), varMap.getSymbolicExpr(Sym.ys))
        );
    }

    EExpr getREZSJoinXYZ(){
        return EEngine.eq(
                varMap.getSymbolicExpr(Sym.rezs),
                EEngine.or(varMap.getSymbolicExpr(Sym.xs), varMap.getSymbolicExpr(Sym.ys), varMap.getSymbolicExpr(Sym.zs))
        );
    }

    private boolean getArithmArgs(){
        x = getXArg();
        y = getYArg();
        if (bytecode.getComputedArguments().size() > 2) {
            if (bytecode.getComputedArguments().get(2) != null) {
                rez = EEngine.eq(
                        varMap.getIntExpr(EVarInt.gp),
                        EEngine.num(bytecode.getComputedArguments().get(2))
                );
                return true;
            }
        }
        return false;
    }

    private boolean getArithmArgsZ(){
        x = getXArg();
        y = getYArg();
        z = getZArg();
        if (bytecode.getComputedArguments().size() > 3) {
            if (bytecode.getComputedArguments().get(3) != null) {
                rez = EEngine.eq(
                        varMap.getIntExpr(EVarInt.gp),
                        EEngine.num(bytecode.getComputedArguments().get(3))
                );
                return true;
            }
        }
        return false;
    }

    private void overapproxBinary() {
            EExpr premState, postmState;

             /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
             );*/

            EExpr rezt = getREZTJoinXY();
            EExpr xt = getXArgT();
            EExpr yt = getYArgT();

            if (memBound) {
                premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd, standardMem);
                postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                        EEngine.storeInt(
                                varMap.getArrExpr(EVarArr.stack),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                varMap.getIntExpr(EVarInt.gp)),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                varMap.getTaintExpr(Taint.rezt)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                EEngine.bool(true)),
                        varMap.getTaintExpr(Taint.taint),
                        cd, standardMem);
            }
            else{
                premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd);
                postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                        EEngine.storeInt(
                                varMap.getArrExpr(EVarArr.stack),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                varMap.getIntExpr(EVarInt.gp)),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                varMap.getTaintExpr(Taint.rezt)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                EEngine.bool(true)),
                        varMap.getTaintExpr(Taint.taint),
                        cd);
                EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                        //, valid
                ), postMem);
                mRules.add(rule2);
            }
            EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
            EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

            EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
            EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

            EExpr rule1 = EEngine.implies(EEngine.and(
                    premState,
                    //valid,
                    xt,
                    yt,
                    rezt
                    ),
                    postmState);
            EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                    //, valid
            ), postgState);
            EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                    //, valid
            ), postStor);
            mRules.add(rule1);
            mRules.add(rule3);
            mRules.add(rule4);
    }

    private void overApproxAppendToStack() {
        EExpr h, b;
        EExpr push1arg = varMap.getIntExpr(EVarInt.va);

        EExpr premState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                varMap.getIntExpr(EVarInt.size));*/

        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            b = pred.mStatePredMem(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            varMap.getIntExpr(EVarInt.size), push1arg),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            h = premState;
                    //,valid);
            mRules.add(EEngine.implies(h, b));
        } else {
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            b = pred.mStatePred(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            varMap.getIntExpr(EVarInt.size), push1arg),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            h = premState;
                    //, valid);
            mRules.add(EEngine.implies(h, b));

            EExpr preMem = pred.memPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            h = EEngine.and(
                    premState,
                    //valid,
                    preMem
            );

            b = pred.memPred(contractID, next,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

            mRules.add(EEngine.implies(h, b));
        }

        EExpr pregState = pred.gStatePred(contractID, pc,
                
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);

        h = EEngine.and(premState,
                //valid,
                pregState);

        b = pred.gStatePred(contractID, next,
                
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));


        EExpr preStor = pred.storPred(contractID, pc,
                
                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        h = EEngine.and(premState,
                //valid,
                preStor);

        b = pred.storPred(contractID, next,
                
                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void valueAppendToStack(EExpr value) {
        EExpr h, b;
        //EExpr push1arg = varMap.getIntExpr(EVarInt.va);

        EExpr premState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                varMap.getIntExpr(EVarInt.size));*/

        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            b = pred.mStatePredMem(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            varMap.getIntExpr(EVarInt.size), value),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(false)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            h = premState;
            //,valid);
            mRules.add(EEngine.implies(h, b));
        } else {
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            b = pred.mStatePred(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            varMap.getIntExpr(EVarInt.size), value),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(false)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            h = premState;
            //, valid);
            mRules.add(EEngine.implies(h, b));

            EExpr preMem = pred.memPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            h = EEngine.and(
                    premState,
                    //valid,
                    preMem
            );

            b = pred.memPred(contractID, next,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

            mRules.add(EEngine.implies(h, b));
        }

        EExpr pregState = pred.gStatePred(contractID, pc,

                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);

        h = EEngine.and(premState,
                //valid,
                pregState);

        b = pred.gStatePred(contractID, next,

                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));


        EExpr preStor = pred.storPred(contractID, pc,

                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        h = EEngine.and(premState,
                //valid,
                preStor);

        b = pred.storPred(contractID, next,

                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void STOP0hornclause() {
        /*EExpr premState;
        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),cd);
        }
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor));
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b));
        EExpr returnPredicate = pred.returnPred(contractID, EEngine.num(0), cd);
        EExpr resgStatePredicate = pred.resgStatePred(contractID, varMap.getIntExpr(EVarInt.aw), varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b));
        EExpr resStor = pred.resStorPred(contractID,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor));
        EExpr rule1 = EEngine.implies(premState, returnPredicate);
        EExpr rule2 = EEngine.implies(EEngine.and(premState, pregState), resgStatePredicate);
        EExpr rule3 = EEngine.implies(EEngine.and(premState, preStor), resStor);

        mRules.add(rule1);
        mRules.add(rule2);
        mRules.add(rule3);*/
    }



    private void ADD1hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.add(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y))
                    )
                    );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
        );*/

        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                     cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs),
                postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void MUL2hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.mul(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y))
                    )
            );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/
        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);

        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs), postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);

    }

    private void SUB3hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.sub(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y))
                    )
            );
        }

        EExpr premState, postmState;

        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
        );*/

        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs), postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void DIV4hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.div(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y))
                    )
            );
        }

        EExpr premState, postmState;

        /*EExpr valid =
                EEngine.checkValid(
                        varMap.getIntExpr(EVarInt.size), EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/

        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            EEngine.ite(EEngine.eq(EEngine.num(0),
                                    varMap.getIntExpr(EVarInt.y)),
                                    EEngine.num(0),
                                    varMap.getIntExpr(EVarInt.gp))),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        } else {
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            EEngine.ite(EEngine.eq(EEngine.num(0),
                                    varMap.getIntExpr(EVarInt.y)),
                                    EEngine.num(0),
                                    varMap.getIntExpr(EVarInt.gp))),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(
                    preMem,
                    premState
                    //, valid
            ), postMem);
            mRules.add(rule2);

        }

        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs), postmState);

        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void SDIV5hornclause() {
        overapproxBinary();
    }

    private void MOD6hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.mod(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y))
                    )
            );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
        );*/
        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs),
                postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);

        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void SMOD7hornclause() {
        overapproxBinary();
    }

    private void ADDMOD8hornclause() {
        EExpr rezt = getREZTJoinXYZ();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();
        EExpr zt = getZArgT();

        EExpr rezs = getREZSJoinXYZ();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();
        EExpr zs = getZArgS();


        if (!getArithmArgsZ()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.mod(
                                    EEngine.add(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y)),
                                    varMap.getIntExpr(EVarInt.z)
                            )                    )
            );
        }

        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
        );*/
        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, zt, rezt, xs, ys, zs, rezs),
                postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void MULMOD9hornclause() {
        EExpr rezt = getREZTJoinXYZ();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();
        EExpr zt = getZArgT();

        EExpr rezs = getREZSJoinXYZ();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();
        EExpr zs = getZArgS();

        if (!getArithmArgsZ()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.mod(
                                    EEngine.mul(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y)),
                                    varMap.getIntExpr(EVarInt.z)
                            )                   )
            );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
        );*/
        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, zt, rezt, xs, ys, zs, rezs),
                postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    //TODO: overaaprox
    private void EXP0ahornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    );
        }

        EExpr h, b;
        EExpr mState;

        if (memBound){
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
        }
        else{
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),  cd);
        }
        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(2))
        );*/

        h = EEngine.and(
                mState,
                //valid,
                x,
                y,
                xt, yt, rez, rezt, xs, ys, rezs
        );

        EExpr checkPremsPred = pred.checkPremsEXPPred(contractID, pc
                , varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y),
                varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                varMap.getTaintExpr(Taint.taint), cd);

        b = checkPremsPred;

        mRules.add(EEngine.implies(h, b));

        h = EEngine.and(mState,
                //valid,
                x,
                y,
                xt, yt, rez, rezt, xs, ys, rezs
        );

        if (memBound) {
            b = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(2)),
                            varMap.getIntExpr(EVarInt.gp)
                    ),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            EEngine.bool(true)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            mRules.add(EEngine.implies(h, b));

        } else {
            b = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(2)),
                            varMap.getIntExpr(EVarInt.gp)),
                            EEngine.storeT(
                                    varMap.getArrTExpr(EVarArrT.stackT),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                    varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            EEngine.bool(true)),
                    varMap.getTaintExpr(Taint.taint)
                    ,
                    cd);
            mRules.add(EEngine.implies(h, b));

            EExpr preMem = pred.memPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

            h = EEngine.and(
                    preMem,
                    checkPremsPred
            );

            b = pred.memPred(contractID, next,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            mRules.add(EEngine.implies(h, b));
        }


        //Propagation of gstate and storage

        EExpr gState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                gState,
                checkPremsPred
        );
        b = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc, 
                varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        h = EEngine.and(
                preStor,
                checkPremsPred
        );
        b = pred.storPred(contractID, next,
                 varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void SIGNEDEXTEND0bhornclause() {
        overapproxBinary();
    }

    private void LT10hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.ite(EEngine.lt(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y)),
                                    EEngine.num(1), EEngine.num(0))                    )
            );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),  EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)), EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint), cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }

        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs), postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void GT11hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.ite(EEngine.gt(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y)),
                                    EEngine.num(1), EEngine.num(0))                 )
            );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)), EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint), cd);
            EExpr preMem = pred.memPred(contractID, pc,  varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }

        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        EExpr preStor = pred.storPred(contractID, pc,   varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,
                 varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs), postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);

    }

    private void SLT12hornclause() {
        overapproxBinary();
    }

    private void SGT13hornclause() {
        overapproxBinary();
    }

    //TODO: overapprox for the moment
    private void EQ14hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        /*if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.ite(EEngine.eq(
                                    varMap.getIntExpr(EVarInt.x),
                                    varMap.getIntExpr(EVarInt.y)
                                    ),
                                    EEngine.num(1), EEngine.num(0)))
            );
        }*/

        EExpr h;

        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/

        EExpr premState, postmState;

        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            EEngine.bool(true)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        } else {
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            EEngine.bool(true)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(
                    preMem,
                    premState
                    //, valid
                    ),
                    postMem);
            mRules.add(rule2);

        }

        h = premState;

        /*h = EEngine.and(premState,
                //valid,
                x,
                y,
                rez,
                xt, yt, rezt, xs, ys, rezs
        );*/

        mRules.add(EEngine.implies(h, postmState));

        EExpr pregState = pred.gStatePred(contractID, pc,   varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,   varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void ISZERO15hornclause() {
        x = getXArg();
        EExpr xt = getXArgT();
        EExpr xs = getXArgS();

        boolean rezDefined = false;
        if (bytecode.getComputedArguments().size() > 0) {
            if (bytecode.getComputedArguments().get(1) != null) {
                rez = EEngine.eq(
                        varMap.getIntExpr(EVarInt.gp),
                        EEngine.num(bytecode.getComputedArguments().get(1))
                );
                rezDefined = true;
            }
        }
        if (!rezDefined){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.xs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),

                    EEngine.eq(
                        varMap.getIntExpr(EVarInt.gp),
                        EEngine.ite(
                            EEngine.eq(varMap.getIntExpr(EVarInt.x), EEngine.num(0)),
                                EEngine.num(1),
                                EEngine.num(0)
                             )
                    )
                    );
        }

        EExpr h, b;

        EExpr mState;

        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1))
        );*/

        if (memBound) {
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            h = EEngine.and(
                    mState,
                    //valid,
                    x,
                    rez,
                    xt,
                    xs
            );
            b = pred.mStatePredMem(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getTaintExpr(Taint.xt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getSymbolicExpr(Sym.xs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            mRules.add(EEngine.implies(h, b));
        } else {
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            h = EEngine.and(
                    mState,
                    //valid,
                    x,
                    rez,
                    xt,
                    xs
            );
            b = pred.mStatePred(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getTaintExpr(Taint.xt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getSymbolicExpr(Sym.xs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            mRules.add(EEngine.implies(h, b));

            // Propagation of remaining memory predicates
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
            h = EEngine.and(
                    mState,
                    //valid,
                    preMem
            );
            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            mRules.add(EEngine.implies(h, b));
        }
        //Propagation of gstate and storage
        EExpr gState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                mState,
                //valid,
                gState
        );
        b = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        h = EEngine.and(
                mState,
                //valid,
                preStor
        );
        b = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void AND16hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.andBitwise(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y))
                    )
            );
        }

        EExpr h, b;
        EExpr mState, postmState;
        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/

        if (memBound) {
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)
                    ),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            h = EEngine.and(
                    mState,
                    x, y, rez, xt, yt, rezt, xs, ys, rezs
            );

            mRules.add(EEngine.implies(h, postmState));
        } else {
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)
                    ),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            h = EEngine.and(
                    mState,
                    //valid,
                    x, y, rez, xt, yt, rezt, xs, ys, rezs
            );

            mRules.add(EEngine.implies(h, postmState));
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

            h = EEngine.and(
                    mState,
                    //valid,
                    preMem
            );

            mRules.add(EEngine.implies(h, postMem));
        }
        EExpr pregState = pred.gStatePred(contractID, pc, 
                varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        EExpr postgState = pred.gStatePred(contractID, next,  
                varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        h = EEngine.and(
                mState,
                //valid,
                pregState
        );

        mRules.add(EEngine.implies(h, postgState));

        EExpr preStor = pred.storPred(contractID, pc, 
                varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,
                 varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        h = EEngine.and(
                mState,
                //valid,
                preStor
        );

        mRules.add(EEngine.implies(h, postStor));
    }

    private void OR17hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.orBitwise(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y))
                    )
            );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/

        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)), EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint), cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }

        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,   varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs), postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void XOR18hornclause() {
        EExpr rezt = getREZTJoinXY();
        EExpr xt = getXArgT();
        EExpr yt = getYArgT();

        EExpr rezs = getREZSJoinXY();
        EExpr xs = getXArgS();
        EExpr ys = getYArgS();

        if (!getArithmArgs()){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.rezs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.xorBitwise(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.y))
                    )
            );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1), EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getIntExpr(EVarInt.gp)), EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getTaintExpr(Taint.rezt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                            varMap.getSymbolicExpr(Sym.rezs)),
                    varMap.getTaintExpr(Taint.taint), cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, y, rez, xt, yt, rezt, xs, ys, rezs), postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void NOT19hornclause() {
        x = getXArg();

        EExpr xt = getXArgT();
        EExpr xs = getXArgS();

        boolean rezDefined = false;
        if (bytecode.getComputedArguments().size() > 0) {
            if (bytecode.getComputedArguments().get(1) != null) {
                rez = EEngine.eq(
                        varMap.getIntExpr(EVarInt.gp),
                        EEngine.num(bytecode.getComputedArguments().get(1))
                );
                rezDefined = true;
            }
        }
        if (!rezDefined){
            rez = EEngine.ite(varMap.getSymbolicExpr(Sym.xs),
                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            varMap.getIntExpr(EVarInt.val0)
                    ),

                    EEngine.eq(
                            varMap.getIntExpr(EVarInt.gp),
                            EEngine.notBitwise(varMap.getIntExpr(EVarInt.x))
                    )
            );
        }
        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1));*/
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getIntExpr(EVarInt.gp)), EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getTaintExpr(Taint.xt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getSymbolicExpr(Sym.xs)),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            postmState = pred.mStatePred(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getIntExpr(EVarInt.gp)), EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getTaintExpr(Taint.xt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getSymbolicExpr(Sym.xs)),
                    varMap.getTaintExpr(Taint.taint), cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState
                    //, valid
            ), postMem);
            mRules.add(rule2);
        }

        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,   varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,   varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState,
                //valid,
                x, rez, xt,
                xs), postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState
                //, valid
        ), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState
                //, valid
        ), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void BYTE1ahornclause(){
        overapproxBinary();
    }

    private void SHA320hornclause() {
        EExpr xval = null;
        if (bytecode.getComputedArguments().size()>0) {
            if (bytecode.getComputedArguments().get(0) != null) {
                xval = EEngine.num(bytecode.getComputedArguments().get(0));
            }
        }

        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/
        if (memBound){
           premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                   varMap.getTaintExpr(Taint.taint), cd, standardMem);
           postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2), varMap.getIntExpr(EVarInt.x)),
                   EEngine.storeT(
                           varMap.getArrTExpr(EVarArrT.stackT),
                           EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                           EEngine.bool(true)),
                   EEngine.storeS(
                           varMap.getArrSExpr(EVarArrS.stackS),
                           EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                           EEngine.bool(true)),
                   varMap.getTaintExpr(Taint.taint), cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),cd);
            if (xval != null){
                postmState = pred.mStatePred(contractID, next,
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 2), xval),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                EEngine.bool(false)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                EEngine.bool(false)),
                        varMap.getTaintExpr(Taint.taint), cd);
            }
            else{
                postmState = pred.mStatePred(contractID, next,
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 2), varMap.getIntExpr(EVarInt.x)),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                EEngine.bool(true)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                EEngine.bool(true)),
                        varMap.getTaintExpr(Taint.taint), cd);
            }

            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va),
                    varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va),
                    varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(premState,
                    //valid,
                    preMem), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(premState, postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(premState, pregState), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(premState, preStor), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void ADDRESS30hornclause(){
        String address = contr.getAddress();
        if (address != null){
            EExpr value = EEngine.num(address);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }
    }

    private void BALANCE31hornclause(){
        String balance = contr.getBalance();
        if (balance != null){
            EExpr value = EEngine.num(balance);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }
    }

    //TODO: at the moment origin is the current contract?
    private void ORIGIN31hornclause() {
        ADDRESS30hornclause();
    }
    private void CALLER33hornclause() {
        ADDRESS30hornclause();
    }

    private void CALLVALUE34hornclause() {
        overApproxAppendToStack();
    }

    private void GASPRICE3ahornclause(){
        String gasPrice = block.getGasPrice();
        if (gasPrice != null){
            EExpr value = EEngine.num(gasPrice);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }

    }

    private void EXTCODESIZE3bhornclause() {
        overApproxAppendToStack();
    }

    private void COINBASE41hornclause() {
        String receiptsRoot = block.getReceiptsRoot();
        if (receiptsRoot != null){
            EExpr value = EEngine.num(receiptsRoot);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }
    }

    private void TIMESTAMP42hornclause() {
        String timestamp = block.getTimestamp();
        if (timestamp != null){
            EExpr value = EEngine.num(timestamp);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }
    }

    private void NUMBER43hornclause() {
        String number = block.getNumber();
        if (number != null){
            EExpr value = EEngine.num(number);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }
    }

    private void DIFFICULTY44hornclause() {
        String difficulty = block.getDifficulty();
        if (difficulty != null){
            EExpr value = EEngine.num(difficulty);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }
    }

    private void GASLIMIT45hornclause() {
        String gasLimit = block.getGasLimit();
        if (gasLimit != null){
            EExpr value = EEngine.num(gasLimit);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }
    }

    private void PC58hornclause() {
        EExpr value = EEngine.num(pc);
        valueAppendToStack(value);
    }

    private void CALLDATALOAD35hornclause() {
        EExpr h, b;
        EExpr mState;

        if (memBound) {
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd,
                    standardMem);
        } else {
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }

        h = mState;
        EExpr postmState;

        if (memBound) {
            postmState = pred.mStatePredMem(contractID, next, varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)), varMap.getIntExpr(EVarInt.v)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        } else {
            postmState = pred.mStatePred(contractID, next, varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)), varMap.getIntExpr(EVarInt.v)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                    cd);
        }

        b = postmState;

        mRules.add(EEngine.implies(h, b));

        // Propagation rules
        EExpr preMem = pred.memPred(contractID, pc,
                varMap.getIntExpr(EVarInt.posmem),
                varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
        h = EEngine.and(
                mState,
                preMem
        );
        b = pred.memPred(contractID, next,
                varMap.getIntExpr(EVarInt.posmem),
                varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

        mRules.add(EEngine.implies(h, b));

        EExpr gState = pred.gStatePred(contractID, pc,

                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                mState,
                gState
        );
        b = pred.gStatePred(contractID, next,

                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc,

                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        h = EEngine.and(
                mState,
                preStor
        );
        b = pred.storPred(contractID, next,

                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        mRules.add(EEngine.implies(h, b));
    }

    private void CALLDATASIZE36GAShornclause() {
        EExpr h, b;

        EExpr mState;

        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                varMap.getIntExpr(EVarInt.size)
        );*/

        if (memBound) {
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);
            h = mState;

            b = pred.mStatePredMem(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), varMap.getIntExpr(EVarInt.size), varMap.getIntExpr(EVarInt.va)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            mRules.add(EEngine.implies(h, b));
        } else {
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);
            h = mState;

            b = pred.mStatePred(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), varMap.getIntExpr(EVarInt.size), varMap.getIntExpr(EVarInt.va)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                    cd);

            mRules.add(EEngine.implies(h, b));

            // Propagation of remaining memory predicates

            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            h = EEngine.and(
                    mState,
                    preMem
            );
            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            mRules.add(EEngine.implies(h, b));
        }

        //Propagation of gstate and storage

        EExpr gState = pred.gStatePred(contractID, pc, varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                mState,
                gState
        );
        b = pred.gStatePred(contractID, next, varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc, varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        h = EEngine.and(
                mState,
                preStor
        );
        b = pred.storPred(contractID, next, varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void CODESIZE38hornclause() {
        EExpr h, b;

        EExpr mState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                varMap.getIntExpr(EVarInt.size)
        );*/

        if (memBound) {
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd, standardMem);


            h = mState;

            b = pred.mStatePredMem(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), varMap.getIntExpr(EVarInt.size), EEngine.num(maxPc + 1)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            mRules.add(EEngine.implies(h, b));
        } else {
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint), cd);


            h = mState;

            b = pred.mStatePred(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), varMap.getIntExpr(EVarInt.size), EEngine.num(maxPc + 1)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                    cd);

            mRules.add(EEngine.implies(h, b));
            // Propagation of remaining memory predicates

            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            h = EEngine.and(
                    mState,
                    preMem
            );
            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            mRules.add(EEngine.implies(h, b));
        }
        //Propagation of gstate and storage

        EExpr gState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                mState,
                gState
        );
        b = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        h = EEngine.and(
                mState,
                preStor
        );
        b = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void CALLDATACOPY37hornclause(){
        CODECOPY39hornclause();
    }

    private void EXTCODECOPY3chornclause() {
        CODECOPY39hornclause();
    }

    private void CODECOPY0CASE(){
        EExpr h, b;

        EExpr premState, postmState;
        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 3));*/
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            h = premState;

            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            b = postmState;
            mRules.add(EEngine.implies(h, b));
        }
        else{
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            h = premState;

            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            b = postmState;
            mRules.add(EEngine.implies(h, b));
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState), postStor);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void CODECOPY0CASECHECK(){
        EExpr h, b;

        EExpr premState, postmState;
        EExpr sIs0 = EEngine.eq(EEngine.num(0),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 3))
        );
        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 3));*/
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            h = EEngine.and(premState, sIs0);

            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            b = postmState;
            mRules.add(EEngine.implies(h, b));
        }
        else{
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            h = EEngine.and(premState, sIs0);

            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                    varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            b = postmState;
            mRules.add(EEngine.implies(h, b));
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState, sIs0), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState, sIs0), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState, sIs0), postStor);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void CODECOPY39hornclause() {
        boolean case0 = false;
        if (bytecode.getComputedArguments().size() > 0) {
            if (bytecode.getComputedArguments().get(2) != null) {
                if (Integer.parseInt(bytecode.getComputedArguments().get(2)) == 0) {
                    CODECOPY0CASE();
                    case0 = true;
                }
            }
        }
        if (!case0){
            CODECOPY0CASECHECK();
            EExpr h, b;
            EExpr premState;

            if (memBound){
                premState = pred.mStatePredMem(contractID, pc,
                        varMap.getIntExpr(EVarInt.size),
                        varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd, standardMem);
            }
            else{
                premState = pred.mStatePred(contractID, pc,
                        varMap.getIntExpr(EVarInt.size),
                        varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd);
            }

            EExpr posMem = EEngine.eq(varMap.getIntExpr(EVarInt.posmem),
                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
            );
            EExpr posCom = EEngine.eq(varMap.getIntExpr(EVarInt.poscom),
                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2))
            );
            EExpr s = EEngine.eq(varMap.getIntExpr(EVarInt.s),
                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 3))
            );

            EExpr noSymbolic = EEngine.and(
                    EEngine.eq(EEngine.bool(false),
                            EEngine.selectS(
                                    varMap.getArrSExpr(EVarArrS.stackS),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1)
                            )
                    ),
                    EEngine.eq(EEngine.bool(false),
                            EEngine.selectS(
                                    varMap.getArrSExpr(EVarArrS.stackS),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
                            )
                    ),
                    EEngine.eq(EEngine.bool(false),
                            EEngine.selectS(
                                    varMap.getArrSExpr(EVarArrS.stackS),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 3)
                            )
                    )
            );

            EExpr symbolic = EEngine.or(
                    EEngine.eq(EEngine.bool(true),
                            EEngine.selectS(
                                    varMap.getArrSExpr(EVarArrS.stackS),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1)
                            )
                    ),
                    EEngine.eq(EEngine.bool(true),
                            EEngine.selectS(
                                    varMap.getArrSExpr(EVarArrS.stackS),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
                            )
                    ),
                    EEngine.eq(EEngine.bool(true),
                            EEngine.selectS(
                                    varMap.getArrSExpr(EVarArrS.stackS),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 3)
                            )
                    )
            );

            /*EExpr valid = EEngine.checkValid(varMap.getIntExpr(EVarInt.size),
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 3));*/

            h = EEngine.and(
                    premState,
                    posMem,
                    posCom,
                    s
            );

            EExpr checkPremsCODECOPY = pred.checkPremsCODECOPYPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.poscom), varMap.getIntExpr(EVarInt.s), varMap.getTaintExpr(Taint.taint), cd);

            b = checkPremsCODECOPY;
            mRules.add(EEngine.implies(h, b));

            //Machine state rule:
            h = premState;

            EExpr postmState;

            if (memBound){
                postmState = pred.mStatePredMem(contractID, next,
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                        varMap.getArrExpr(EVarArr.stack),
                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd, ovappxMem);
                b = postmState;
                mRules.add(EEngine.implies(h, b));
            }
            else{
                postmState = pred.mStatePred(contractID, next,
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 3),
                        varMap.getArrExpr(EVarArr.stack),
                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd);
                b = postmState;
                mRules.add(EEngine.implies(h, b));


                EExpr WmemMod32 = EEngine.eq(EEngine.num(0), EEngine.mod(varMap.getIntExpr(EVarInt.wmem), EEngine.num(32)));


                EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.wmem), varMap.getIntExpr(EVarInt.va),
                        varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                EExpr postMem = pred.memPred(contractID, next,
                        varMap.getIntExpr(EVarInt.wmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                EExpr postMemOver = pred.memPred(contractID, next,
                        varMap.getIntExpr(EVarInt.wmem), varMap.getIntExpr(EVarInt.val0), EEngine.bool(true), EEngine.bool(true), cd);

                //Overapproximation if symbolic

                mRules.add(EEngine.implies(EEngine.and(
                        premState,
                        symbolic
                ), postMemOver));

                //Overapproximation for the updated memory
                EExpr WmemGE = EEngine.ge(varMap.getIntExpr(EVarInt.wmem), EEngine.sub(varMap.getIntExpr(EVarInt.posmem), EEngine.num(31)));
                EExpr WmemLT =
                        EEngine.lt(varMap.getIntExpr(EVarInt.wmem), EEngine.add(varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.s)));

                mRules.add(EEngine.implies(EEngine.and(
                        premState,
                        posMem,
                        posCom,
                        s,
                        noSymbolic,
                        WmemGE,
                        WmemLT,
                        WmemMod32
                ), postMemOver));

                //Propagation rules for remaining memory

                //Memory propagation (low)
                EExpr WmemPosmem31 = EEngine.lt(varMap.getIntExpr(EVarInt.wmem), EEngine.sub(varMap.getIntExpr(EVarInt.posmem), EEngine.num(31)));
                mRules.add(EEngine.implies(EEngine.and(
                        premState,
                        posMem,
                        posCom,
                        s,
                        noSymbolic,
                        WmemPosmem31,
                        WmemMod32,
                        preMem
                ), postMem));

                //Memory propagation (high)
                EExpr WmemPosmemS =
                        EEngine.ge(varMap.getIntExpr(EVarInt.wmem), EEngine.add(varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.s)));
                mRules.add(EEngine.implies(EEngine.and(
                        premState,
                        posMem,
                        posCom,
                        s,
                        noSymbolic,
                        WmemPosmemS,
                        WmemMod32,
                        preMem
                ), postMem));
            }

            //Global state propagation (GState)
            EExpr pregState = pred.gStatePred(contractID, pc, 
                    varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
            EExpr postgState = pred.gStatePred(contractID, next,
                     varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
            mRules.add(EEngine.implies(EEngine.and(checkPremsCODECOPY, pregState), postgState));

            //Global state propagation (global storage)
            EExpr preStor = pred.storPred(contractID, pc, 
                    varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
            EExpr postStor = pred.storPred(contractID, next,
                     varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
            mRules.add(EEngine.implies(EEngine.and(checkPremsCODECOPY, preStor), postStor));
        }
    }

    private void BLOCKHASH40hornclause() {
        String hash = block.getHash();
        if (hash != null){
            EExpr value = EEngine.num(hash);
            valueAppendToStack(value);
        }
        else{
            overApproxAppendToStack();
        }
        /*EExpr premState, postmState;
        //EExpr valid = EEngine.checkValid(varMap.getIntExpr(EVarInt.size), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1));
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1), varMap.getIntExpr(EVarInt.v)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            postmState = pred.mStatePred(contractID, next, varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1), varMap.getIntExpr(EVarInt.v)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(premState, postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);*/
    }

    private void POP50hornclause() {
        EExpr premState, postmState;
        //EExpr valid = EEngine.checkValid(varMap.getIntExpr(EVarInt.size), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1));
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            postmState = pred.mStatePred(contractID, next, EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule4 = EEngine.implies(EEngine.and(preMem, premState), postMem);
            mRules.add(rule4);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(premState, postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(premState, pregState), postgState);
        EExpr rule5 = EEngine.implies(EEngine.and(preStor, premState), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule5);
    }

//    private void MLOAD0CASE(){
//        EExpr h, b;
//        EExpr mState;
//        EExpr maEqSa = EEngine.eq(
//                varMap.getIntExpr(EVarInt.ma),
//                EEngine.select(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)))
//        );
//
//        EExpr checkPrems = pred.checkPremsMLOADPred(contractID, pc,
//                varMap.getIntExpr(EVarInt.ma), cd
//                );
//        if (memBound) {
//            mState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), cd, standardMem);
//            for (int i = 0; i < memSize; i++) {
//                if (i % 32 != 0) continue;
//                h = EEngine.and(
//                        checkPrems,
//                        mState,
//                        maEqSa,
//                        EEngine.eq(
//                                EEngine.num(i),
//                                varMap.getIntExpr(EVarInt.ma)
//                        )
//                );
//
//                b = pred.mStatePredMem(contractID, next,
//                        varMap.getIntExpr(EVarInt.size),
//                        EEngine.store(varMap.getArrExpr(EVarArr.stack),
//                                EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
//                                varMap.getVal(i)
//                        ),
//                        cd, standardMem);
//
//                mRules.add(EEngine.implies(h, b));
//                //TODO: error case
//        /*h = EEngine.and(
//                checkPrems,
//                mState,
//                maEqSa,
//                EEngine.gt(
//                        varMap.getIntExpr(EVarInt.ma),
//                        EEngine.num(memSize)
//                )
//                //,memWord
//        );
//
//        b = pred.mStatePred(contractID, next,
//                varMap.getIntExpr(EVarInt.size),
//                EEngine.store(varMap.getArrExpr(EVarArr.stack),
//                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
//                        varMap.getVal(i)
//                ),
//                varMap.getIntExpr(EVarInt.awnew), standardMem);
//
//        mRules.add(EEngine.implies(h, b));*/
//            }
//        } else {
//            mState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), cd);
//            EExpr memWord = pred.memPred(contractID, pc,
//                    varMap.getIntExpr(EVarInt.ma), varMap.getIntExpr(EVarInt.val0), cd);
//            h = EEngine.and(
//                    checkPrems,
//                    mState,
//                    maEqSa
//                    , memWord
//            );
//
//            b = pred.mStatePred(contractID, next,
//                    varMap.getIntExpr(EVarInt.size),
//                    EEngine.store(varMap.getArrExpr(EVarArr.stack),
//                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
//                            varMap.getIntExpr(EVarInt.val0)
//                    )
//                    , cd);
//
//            mRules.add(EEngine.implies(h, b));
//
//            // Propagation of remaining memory predicates
//
//            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), cd);
//            h = EEngine.and(
//                    checkPrems,
//                    preMem
//            );
//            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), cd);
//            mRules.add(EEngine.implies(h, b));
//            //TODO: error case
//            /*
//            h = EEngine.and(
//                    checkPrems,
//                    mState,
//                    maEqSa,
//                    EEngine.gt(
//                            varMap.getIntExpr(EVarInt.ma),
//                            EEngine.num(memSize)
//                    )
//                    ,memWord
//            );
//
//            b = pred.mStatePred(contractID, next,
//                    varMap.getIntExpr(EVarInt.size),
//                    EEngine.store(varMap.getArrExpr(EVarArr.stack),
//                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
//                            varMap.getVal(i)
//                    ),
//                    varMap.getIntExpr(EVarInt.awnew), standardMem);
//
//            mRules.add(EEngine.implies(h, b));*/
//        }
//    }


    private void MLOAD51hornclause() {
        EExpr h, b;

        EExpr mState;

        if (memBound) {
            mState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd,
                    standardMem);

        } else {
            mState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }

        Boolean mod32is0 = null;
        String stringMA = null;
        if (bytecode.getComputedArguments().size() > 1) {
            if (bytecode.getComputedArguments().get(0) != null) {
                stringMA = bytecode.getComputedArguments().get(0);
                BigInteger bigMA = new BigInteger(stringMA);
                BigInteger modRez = bigMA.mod(BigInteger.valueOf(32));
                if (modRez.compareTo(BigInteger.valueOf(0)) == 0){
                    mod32is0 = true;
                }
                else{
                    mod32is0 = false;
                }
            }
        }
        String stringVAL = null;
        if (bytecode.getComputedArguments().size() > 2) {
            if (bytecode.getComputedArguments().get(1) != null) {
                stringVAL = bytecode.getComputedArguments().get(1);
            }
        }

        EExpr maEqSa = EEngine.eq(
                varMap.getIntExpr(EVarInt.ma),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)))
        );

        EExpr maIsSym = EEngine.eq(
                EEngine.bool(true),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        EExpr maIsNotSym = EEngine.eq(
                EEngine.bool(false),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1));*/


        h = EEngine.and(
                mState,
                maEqSa
                //valid
        );

        EExpr checkPrems = pred.checkPremsMLOADPred(contractID, pc,
                varMap.getIntExpr(EVarInt.ma), varMap.getTaintExpr(Taint.taint), cd
                );


        b = checkPrems;

        mRules.add(EEngine.implies(h, b));

        /*if (bytecode.getComputedArguments().size() > 0) {
            if (bytecode.getComputedArguments().get(0) != null) {
                if ((Integer.parseInt(bytecode.getComputedArguments().get(0)) % 32) == 0) {
                    MLOAD0CASE();
                }
            }
        } else {*/

            EExpr maMod32Eq0 = EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.ma), EEngine.num(32)), EEngine.num(0));
            if (memBound) {
                // Case 1
                for (int i = 0; i < memSize; i++) {
                    if (i % 32 != 0) continue;
                    h = EEngine.and(
                            maEqSa,
                            mState,
                            maMod32Eq0,
                            EEngine.eq(
                                    EEngine.num(i),
                                    varMap.getIntExpr(EVarInt.ma)
                            )
                    );

                    b = pred.mStatePredMem(contractID, next,
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                                    varMap.getVal(i)
                            ),
                            EEngine.storeT(
                                    varMap.getArrTExpr(EVarArrT.stackT),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                    EEngine.bool(true)),
                            EEngine.storeS(
                                    varMap.getArrSExpr(EVarArrS.stackS),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                    EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                            cd, standardMem);

                    mRules.add(EEngine.implies(h, b));
                }

                // Case 2
                EExpr notMaMod32Eq0 = EEngine.not(EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.ma), EEngine.num(32)), EEngine.num(0)));

                h = EEngine.and(
                        mState,
                        maEqSa,
                        notMaMod32Eq0
                );

                b = pred.mStatePredMem(contractID, next,
                        varMap.getIntExpr(EVarInt.size),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)), varMap.getIntExpr(EVarInt.vnew)),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                EEngine.bool(true)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                        cd, standardMem);

                mRules.add(EEngine.implies(h, b));


            } else {
                // Case 1

                EExpr memWord = pred.memPred(contractID, pc,
                        varMap.getIntExpr(EVarInt.ma), varMap.getIntExpr(EVarInt.val0), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

                EExpr maEqStringMA = EEngine.eq(
                        varMap.getIntExpr(EVarInt.ma),
                        EEngine.num(stringMA)
                );

                if (mod32is0 != null){
                    if (mod32is0) {
                        if (stringVAL != null){
                            h = mState;
                            b = pred.mStatePred(contractID, next,
                                    varMap.getIntExpr(EVarInt.size),
                                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                                            EEngine.num(stringVAL)
                                    ),
                                    EEngine.storeT(
                                            varMap.getArrTExpr(EVarArrT.stackT),
                                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                            EEngine.bool(false)),
                                    EEngine.storeS(
                                            varMap.getArrSExpr(EVarArrS.stackS),
                                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                                    cd);

                            mRules.add(EEngine.implies(h, b));
                        }
                        else {
                            h = EEngine.and(
                                    mState,
                                    maEqStringMA
                                    , memWord
                            );
                            b = pred.mStatePred(contractID, next,
                                    varMap.getIntExpr(EVarInt.size),
                                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                                            varMap.getIntExpr(EVarInt.val0)
                                    ),
                                    EEngine.storeT(
                                            varMap.getArrTExpr(EVarArrT.stackT),
                                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                            varMap.getTaintExpr(Taint.taintM)),
                                    EEngine.storeS(
                                            varMap.getArrSExpr(EVarArrS.stackS),
                                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                            varMap.getSymbolicExpr(Sym.symM)), varMap.getTaintExpr(Taint.taint),
                                    cd);

                            mRules.add(EEngine.implies(h, b));
                        }
                    }
                    else{
                        h = mState;
                        b = pred.mStatePred(contractID, next,
                                varMap.getIntExpr(EVarInt.size),
                                EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)), varMap.getIntExpr(EVarInt.vnew)),
                                EEngine.storeT(
                                        varMap.getArrTExpr(EVarArrT.stackT),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                        EEngine.bool(true)),
                                EEngine.storeS(
                                        varMap.getArrSExpr(EVarArrS.stackS),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                        EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                                cd);

                        mRules.add(EEngine.implies(h, b));
                    }

                }
                else{
                    //Case 1
                    if (stringVAL != null){
                        h = EEngine.and(
                                mState,
                                maEqSa,
                                maIsNotSym,
                                maMod32Eq0
                        );

                        b = pred.mStatePred(contractID, next,
                                varMap.getIntExpr(EVarInt.size),
                                EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                                        EEngine.num(stringVAL)
                                ),
                                EEngine.storeT(
                                        varMap.getArrTExpr(EVarArrT.stackT),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                        EEngine.bool(false)),
                                EEngine.storeS(
                                        varMap.getArrSExpr(EVarArrS.stackS),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                        EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                                cd);

                        mRules.add(EEngine.implies(h, b));
                    }
                    else {
                        h = EEngine.and(
                                mState,
                                maEqSa,
                                maIsNotSym,
                                maMod32Eq0
                                , memWord
                        );

                        b = pred.mStatePred(contractID, next,
                                varMap.getIntExpr(EVarInt.size),
                                EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                                        varMap.getIntExpr(EVarInt.val0)
                                ),
                                EEngine.storeT(
                                        varMap.getArrTExpr(EVarArrT.stackT),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                        varMap.getTaintExpr(Taint.taintM)),
                                EEngine.storeS(
                                        varMap.getArrSExpr(EVarArrS.stackS),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                        varMap.getSymbolicExpr(Sym.symM)), varMap.getTaintExpr(Taint.taint),
                                cd);

                        mRules.add(EEngine.implies(h, b));

                        h = EEngine.and(
                                mState,
                                maIsSym
                        );

                        b = pred.mStatePred(contractID, next,
                                varMap.getIntExpr(EVarInt.size),
                                EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                                        varMap.getIntExpr(EVarInt.vnew)
                                ),
                                EEngine.storeT(
                                        varMap.getArrTExpr(EVarArrT.stackT),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                        EEngine.bool(true)),
                                EEngine.storeS(
                                        varMap.getArrSExpr(EVarArrS.stackS),
                                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                        EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                                cd);

                        mRules.add(EEngine.implies(h, b));
                    }
                    //Case 2
                    EExpr notMaMod32Eq0 = EEngine.not(EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.ma), EEngine.num(32)), EEngine.num(0)));
                    h = EEngine.and(
                            mState,
                            maEqSa,
                            maIsNotSym,
                            notMaMod32Eq0
                            //,memWord
                    );

                    b = pred.mStatePred(contractID, next,
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)), varMap.getIntExpr(EVarInt.vnew)),
                            EEngine.storeT(
                                    varMap.getArrTExpr(EVarArrT.stackT),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                    EEngine.bool(true)),
                            EEngine.storeS(
                                    varMap.getArrSExpr(EVarArrS.stackS),
                                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                    EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                            cd);

                    mRules.add(EEngine.implies(h, b));
                }

                // Propagation of remaining memory predicates

                EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                h = EEngine.and(
                        checkPrems,
                        preMem
                );
                b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                mRules.add(EEngine.implies(h, b));
            }

        EExpr gState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                checkPrems,
                gState
        );
        b = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor),
                varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        h = EEngine.and(
                checkPrems,
                preStor
        );
        b = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor),
                varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

//    private void MSTORE0CASE(){
//        EExpr h, b;
//        EExpr mState;
//        if (memBound){
//            mState = pred.mStatePredMem(contractID, pc,
//                    varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), cd, standardMem);
//            EExpr checkPrems = pred.checkPremsMSTOREPred(contractID, pc, varMap.getIntExpr(EVarInt.ma),
//                    varMap.getIntExpr(EVarInt.va), cd);
//            for (int i = 0; i < memSize; i++) {
//                if (i % 32 != 0) continue;
//                h = EEngine.and(
//                        mState,
//                        checkPrems,
//                        EEngine.eq(varMap.getIntExpr(EVarInt.ma), EEngine.num(i))
//                );
//                Map<Integer, EExpr> memUp = new HashMap<>();
//                memUp.put(i, varMap.getIntExpr(EVarInt.va));
//                b = pred.mStatePredMem(contractID, next,
//                        varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),cd, varMap.getMemUp(memUp));
//                mRules.add(EEngine.implies(h, b));
//            }
//        } else {
//            mState = pred.mStatePred(contractID, pc,
//                    varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), cd);
//            EExpr checkPrems = pred.checkPremsMSTOREPred(contractID, pc, varMap.getIntExpr(EVarInt.ma),
//                    varMap.getIntExpr(EVarInt.va), cd);
//            h = EEngine.and(
//                    mState,
//                    checkPrems
//            );
//            b = pred.mStatePred(contractID, next,
//                    varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), cd);
//            mRules.add(EEngine.implies(h, b));
//
//
//            // Case 1
//            EExpr postMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.ma), varMap.getIntExpr(EVarInt.va), cd);
//
//            h = EEngine.and(
//                    mState,
//                    checkPrems
//            );
//
//            b = postMem;
//            mRules.add(EEngine.implies(h, b));
//
//            // Propagation of remaining memory predicates
//
//            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), cd);
//            h = EEngine.and(
//                    checkPrems,
//                    preMem,
//                    EEngine.eq(
//                            EEngine.mod(varMap.getIntExpr(EVarInt.posmem), EEngine.num(32)),
//                            EEngine.num(0)
//                    ),
//                    EEngine.le(
//                            varMap.getIntExpr(EVarInt.posmem),
//                            EEngine.sub(varMap.getIntExpr(EVarInt.ma), EEngine.num(32))
//                    )
//            );
//            b = pred.memPred(contractID, next,
//                    varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), cd);
//            mRules.add(EEngine.implies(h, b));
//
//            h = EEngine.and(
//                    checkPrems,
//                    preMem,
//                    EEngine.eq(
//                            EEngine.mod(varMap.getIntExpr(EVarInt.posmem), EEngine.num(32)),
//                            EEngine.num(0)
//                    ),
//                    EEngine.ge(
//                            varMap.getIntExpr(EVarInt.posmem),
//                            EEngine.add(varMap.getIntExpr(EVarInt.ma), EEngine.num(32))
//                    )
//            );
//
//            b = pred.memPred(contractID, next,
//                    varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), cd);
//            mRules.add(EEngine.implies(h, b));
//            }
//    }

    private void MSTORE853hornclause() {
        MSTORE52hornclause();
    }

    private void MSTORE52hornclause() {

        Boolean mod32is0 = null;
        String stringMA = null;
        if (bytecode.getComputedArguments().size() > 1) {
            if (bytecode.getComputedArguments().get(0) != null) {
                stringMA = bytecode.getComputedArguments().get(0);
                BigInteger bigMA = new BigInteger(stringMA);
                BigInteger modRez = bigMA.mod(BigInteger.valueOf(32));
                if (modRez.compareTo(BigInteger.valueOf(0)) == 0){
                    mod32is0 = true;
                }
                else{
                    mod32is0 = false;
                }
            }
        }

        String stringVAL = null;
        if (bytecode.getComputedArguments().size() > 2) {
            if (bytecode.getComputedArguments().get(1) != null) {
                stringVAL = bytecode.getComputedArguments().get(1);
            }
        }

        EExpr maIsSym = EEngine.eq(
                EEngine.bool(true),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        EExpr maIsNotSym = EEngine.eq(
                EEngine.bool(false),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        EExpr h, b;

        EExpr mState;

        if (memBound){
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }

        EExpr maEqSa = EEngine.eq(
                varMap.getIntExpr(EVarInt.ma),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)))
        );

        EExpr vEqSa = EEngine.eq(
                varMap.getIntExpr(EVarInt.va),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(2)))
        );

        EExpr yt = getYArgT();
        EExpr ys = getYArgS();

        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2));*/

        h = EEngine.and(
                mState,
                maEqSa,
                vEqSa,
                yt,
                ys
        );

        EExpr checkPrems = pred.checkPremsMSTOREPred(contractID, pc, varMap.getIntExpr(EVarInt.ma),
                varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.yt), varMap.getSymbolicExpr(Sym.ys), varMap.getTaintExpr(Taint.taint), cd);

        b = checkPrems;

        mRules.add(EEngine.implies(h, b));

        /*if (bytecode.getComputedArguments().size() > 0) {
            if (bytecode.getComputedArguments().get(0) != null) {
                if ((Integer.parseInt(bytecode.getComputedArguments().get(0)) % 32) == 0) {
                   MSTORE0CASE();
                }
            }
        }
        else {*/
            if (memBound) {
                // Case 1
                for (int i = 0; i < memSize; i++) {
                    if (i % 32 != 0) continue;
                    h = EEngine.and(
                            mState,
                            maEqSa,
                            vEqSa,
                            yt,
                            ys
                            , EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.ma), EEngine.num(32)), EEngine.num(0)),
                            EEngine.eq(varMap.getIntExpr(EVarInt.ma), EEngine.num(i))
                    );

                    Map<Integer, EExpr> memUp = new HashMap<>();
                    memUp.put(i, varMap.getIntExpr(EVarInt.va));
                    b = pred.mStatePredMem(contractID, next,
                            varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                            varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getTaintExpr(Taint.taint),
                            cd, varMap.getMemUp(memUp));
                    mRules.add(EEngine.implies(h, b));
                }

                //Case 2
                EExpr notMaMod32Eq0 = EEngine.not(EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.ma), EEngine.num(32)), EEngine.num(0)));
                h = EEngine.and(
                        mState,
                        maEqSa,
                        vEqSa,
                        yt,
                        ys,
                        notMaMod32Eq0
                );
                b = pred.mStatePredMem(contractID, next,
                        varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd, ovappxMem);

                mRules.add(EEngine.implies(h, b));
            } else {
                h = mState;
                b = pred.mStatePred(contractID, next,
                        varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd);
                mRules.add(EEngine.implies(h, b));

                // Case 1

                EExpr postMemUp = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.ma), varMap.getIntExpr(EVarInt.va),
                        varMap.getTaintExpr(Taint.yt), varMap.getSymbolicExpr(Sym.ys), cd);
                EExpr postMemOverUp = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.ma), varMap.getIntExpr(EVarInt.val0),
                        varMap.getTaintExpr(Taint.yt), EEngine.bool(true),
                        cd);
                EExpr maEqStringMA = EEngine.eq(
                        varMap.getIntExpr(EVarInt.ma),
                        EEngine.num(stringMA)
                );

                if (mod32is0 != null){
                    if (mod32is0){
                        if (stringVAL != null){
                            h = EEngine.and(
                                    mState,
                                    maEqStringMA
                            );
                            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.ma), EEngine.num(stringVAL),
                                    EEngine.bool(false), EEngine.bool(false), cd);
                            mRules.add(EEngine.implies(h, b));
                        }
                        else {
                            h = EEngine.and(
                                    mState,
                                    maEqStringMA,
                                    vEqSa,
                                    yt,
                                    ys
                                    //,preMemUp
                            );
                            b = postMemUp;
                            mRules.add(EEngine.implies(h, b));
                        }

                        //Propagation
                        EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
                        h = EEngine.and(
                                mState,
                                maEqStringMA,
                                vEqSa,
                                //valid,
                                preMem,
                                EEngine.not(EEngine.eq(
                                        varMap.getIntExpr(EVarInt.posmem),
                                        varMap.getIntExpr(EVarInt.ma)
                                ))
                        );
                        b = pred.memPred(contractID, next,
                                varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
                        mRules.add(EEngine.implies(h, b));

                    }
                    else{
                        h = EEngine.and(
                                mState,
                                maEqStringMA,
                                vEqSa
                                //,preMemUp
                        );
                        b = postMemOverUp;

                        mRules.add(EEngine.implies(h, b));
                        //Propagation
                        EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
                        h = EEngine.and(
                                mState,
                                maEqStringMA,
                                vEqSa,
                                //valid,
                                preMem,
                                EEngine.not(EEngine.eq(
                                        varMap.getIntExpr(EVarInt.posmem),
                                        varMap.getIntExpr(EVarInt.ma)
                                ))
                        );
                        b = pred.memPred(contractID, next,
                                varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
                        mRules.add(EEngine.implies(h, b));
                    }
                }
                else{
                    //Case 1
                    if (stringVAL != null){
                        h = EEngine.and(
                                mState,
                                maEqSa,
                                vEqSa,
                                maIsNotSym
                                , EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.ma), EEngine.num(32)), EEngine.num(0))
                                //,preMemUp
                        );
                        b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.ma), EEngine.num(stringVAL),
                                EEngine.bool(false), EEngine.bool(false), cd);
                        mRules.add(EEngine.implies(h, b));
                    }
                    else {
                        h = EEngine.and(
                                mState,
                                maEqSa,
                                vEqSa,
                                yt,
                                ys,
                                maIsNotSym
                                , EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.ma), EEngine.num(32)), EEngine.num(0))
                                //,preMemUp
                        );
                        b = postMemUp;
                        mRules.add(EEngine.implies(h, b));
                    }
                    //Case 2
                    EExpr notMaMod32Eq0 = EEngine.not(EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.ma), EEngine.num(32)), EEngine.num(0)));
                    h = EEngine.and(
                            mState,
                            maEqSa,
                            vEqSa,
                            maIsNotSym,
                            notMaMod32Eq0
                            //, preMemUp
                    );
                    b = postMemOverUp;
                    mRules.add(EEngine.implies(h, b));
                    //Propagation of the remaining memory predicates
                    EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
                    h = EEngine.and(
                            mState,
                            maEqSa,
                            vEqSa,
                            //valid,
                            preMem,
                            maIsNotSym,
                            EEngine.not(EEngine.eq(
                                    varMap.getIntExpr(EVarInt.posmem),
                                    varMap.getIntExpr(EVarInt.ma)
                            ))
                    );
                    b = pred.memPred(contractID, next,
                            varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
                    mRules.add(EEngine.implies(h, b));

                    h = EEngine.and(
                            mState,
                            maEqSa,
                            vEqSa,
                            //valid,
                            preMem,
                            maIsSym
                    );
                    b = pred.memPred(contractID, next,
                            varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
                    mRules.add(EEngine.implies(h, b));
                }
            }
        //Propagation of gstate and storage

        EExpr gState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                checkPrems,
                gState
        );
        b = pred.gStatePred(contractID, next, 
                varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor),
                varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        h = EEngine.and(
                checkPrems,
                preStor
        );
        b = pred.storPred(contractID, next,
                 varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor),
                varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void SLOAD54hornclause() {
        String stringMA = null;
        if (bytecode.getComputedArguments().size() > 1) {
            if (bytecode.getComputedArguments().get(0) != null) {
                stringMA = bytecode.getComputedArguments().get(0);
            }
        }

        String stringVAL = null;
        if (bytecode.getComputedArguments().size() > 2) {
            if (bytecode.getComputedArguments().get(1) != null) {
                stringVAL = bytecode.getComputedArguments().get(1);
            }
        }

        EExpr h, b;
        EExpr premState;

        if (memBound){
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }
        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1));*/

        /*EExpr exenvPredicate = pred.exenvPred(
                contractID,
                
                varMap.getIntExpr(EVarInt.i),
                cd,
                varMap.getIntExpr(EVarInt.va),
                varMap.getIntExpr(EVarInt.inputsize)
                );*/

        EExpr preStorLOAD = pred.storPred(contractID, pc,
                varMap.getIntExpr(EVarInt.ma),
                varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.yt), varMap.getSymbolicExpr(Sym.ys), cd);

        EExpr ma;
        if (stringMA != null){
            ma = EEngine.eq(varMap.getIntExpr(EVarInt.ma),
                    EEngine.num(stringMA));
        }
        else{
            ma = EEngine.eq(varMap.getIntExpr(EVarInt.ma),
                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1)));
        }


        EExpr maIsSym = EEngine.eq(
                EEngine.bool(true),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        EExpr maIsNotSym = EEngine.eq(
                EEngine.bool(false),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        if (stringMA != null){
            if (stringVAL != null){
                h = EEngine.and(
                        premState,
                        ma
                );
            }
            else{
                h = EEngine.and(
                        premState,
                        preStorLOAD,
                        ma
                );
            }
        }
        else{
            h = EEngine.and(
                    premState,
                    maIsNotSym,
                    preStorLOAD,
                    ma
            );
        }

        if (memBound){
            b = pred.mStatePredMem(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1), varMap.getIntExpr(EVarInt.v)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getTaintExpr(Taint.yt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                            varMap.getSymbolicExpr(Sym.ys)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            mRules.add(EEngine.implies(h, b));

        } else {
            if (stringVAL != null){
                b = pred.mStatePred(contractID, next,
                        varMap.getIntExpr(EVarInt.size),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1), EEngine.num(stringVAL)),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                EEngine.bool(false)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                        cd);
            }
            else{
                b = pred.mStatePred(contractID, next,
                        varMap.getIntExpr(EVarInt.size),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1), varMap.getIntExpr(EVarInt.v)),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                varMap.getTaintExpr(Taint.yt)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                varMap.getSymbolicExpr(Sym.ys)), varMap.getTaintExpr(Taint.taint),
                        cd);
            }

            mRules.add(EEngine.implies(h, b));

            if (stringMA == null) {
                h = EEngine.and(
                        premState,
                        maIsSym
                );
                b = pred.mStatePred(contractID, next,
                        varMap.getIntExpr(EVarInt.size),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1), varMap.getIntExpr(EVarInt.v)),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                EEngine.bool(true)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                        cd);
                mRules.add(EEngine.implies(h, b));
            }

            EExpr preMem = pred.memPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.pos),
                    varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

            h = EEngine.and(
                    premState,
                    preMem
            );

            b = pred.memPred(contractID, next,
                    varMap.getIntExpr(EVarInt.pos),
                    varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

            mRules.add(EEngine.implies(h, b));
        }

        EExpr pregState = pred.gStatePred(contractID, pc,
                
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);

        h = EEngine.and(
                premState,
                pregState
        );

        b = pred.gStatePred(contractID, next,
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);

        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc,
                
                varMap.getIntExpr(EVarInt.pos),
                varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        h = EEngine.and(
                premState,
                preStor
        );

        b = pred.storPred(contractID, next,
                
                varMap.getIntExpr(EVarInt.pos),
                varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        mRules.add(EEngine.implies(h, b));
    }

    private void SSTORE55hornclause() {
        String stringMA = null;
        if (bytecode.getComputedArguments().size() > 1) {
            if (bytecode.getComputedArguments().get(0) != null) {
                stringMA = bytecode.getComputedArguments().get(0);
            }
        }

        String stringVAL = null;
        if (bytecode.getComputedArguments().size() > 2) {
            if (bytecode.getComputedArguments().get(1) != null) {
                stringVAL = bytecode.getComputedArguments().get(1);
            }
        }

        EExpr h, b;
        EExpr premState, postmState;

        EExpr ma;

        if (stringMA != null){
            ma = EEngine.eq(varMap.getIntExpr(EVarInt.ma), EEngine.num(stringMA));
        }
        else{
            ma = EEngine.eq(varMap.getIntExpr(EVarInt.ma), EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1)));
        }
        EExpr y = EEngine.eq(varMap.getIntExpr(EVarInt.y), EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)));
        EExpr yt = getYArgT();
        EExpr ys = getYArgS();

        //EExpr preStorMa = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.ma), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        /*EExpr checkPremsSSTOREPredicate;
        if (stringVAL != null){
            checkPremsSSTOREPredicate = pred.checkPremsSSTOREPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.ma),
                    EEngine.num(stringVAL),
                    EEngine.bool(false),
                    EEngine.bool(false),
                    varMap.getTaintExpr(Taint.taint), cd);
        }
        else{
            checkPremsSSTOREPredicate = pred.checkPremsSSTOREPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.ma),
                    varMap.getIntExpr(EVarInt.y),
                    varMap.getTaintExpr(Taint.yt),
                    varMap.getSymbolicExpr(Sym.ys),
                    varMap.getTaintExpr(Taint.taint), cd);
        }*/

        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            /*
            if (stringVAL != null){
                h = EEngine.and(
                        premState,
                        ma
                );
            }
            else{
                h = EEngine.and(
                        premState,
                        ma,
                        y,
                        yt,
                        ys
                );
            }

            mRules.add(EEngine.implies(h, checkPremsSSTOREPredicate));*/
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            /*
            if (stringVAL != null){
                h = EEngine.and(
                        premState,
                        ma
                );
            }
            else{
                h = EEngine.and(
                        premState,
                        ma,
                        y,
                        yt,
                        ys
                );
            }

            mRules.add(EEngine.implies(h, checkPremsSSTOREPredicate));*/

            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va),
                    varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va),
                    varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            h = EEngine.and(premState, preMem);
            b = postMem;
            mRules.add(EEngine.implies(h, b));
        }

        if (memBound){
            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 2), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd,
                    standardMem);
        }
        else{
            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 2), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }

        h = premState;
        b = postmState;
        mRules.add(EEngine.implies(h, b));

        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(premState, pregState);
        mRules.add(EEngine.implies(h, postgState));

        EExpr postStorSSTORE;
        if (stringVAL != null){
            postStorSSTORE = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.ma), EEngine.num(stringVAL),  EEngine.bool(false),
                    EEngine.bool(false), cd);
        }
        else{
            postStorSSTORE = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.ma), varMap.getIntExpr(EVarInt.y),  varMap.getTaintExpr(Taint.yt),
                    varMap.getSymbolicExpr(Sym.ys), cd);
        }

        if (stringVAL != null){
            h = EEngine.and(
                    premState,
                    ma
            );
        }
        else{
            h = EEngine.and(
                    premState,
                    ma,
                    y,
                    yt,
                    ys
            );
        }

        mRules.add(EEngine.implies(
                h,
                postStorSSTORE)
        );

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor),
                varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS),  cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor),
                varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS),  cd);
        EExpr maNeqpos = EEngine.not(EEngine.eq(varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.ma)));

        if (stringMA != null){
            h = EEngine.and(
                    premState,
                    ma,
                    preStor,
                    maNeqpos
            );
            mRules.add(EEngine.implies(h, postStor));
        }
        else{
            EExpr maIsNotSym = EEngine.eq(
                    EEngine.bool(false),
                    EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
            );

            h = EEngine.and(
                    premState,
                    ma,
                    maIsNotSym,
                    preStor,
                    maNeqpos
            );

            mRules.add(EEngine.implies(h, postStor));

            EExpr maIsSym = EEngine.eq(
                    EEngine.bool(true),
                    EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
            );

            h = EEngine.and(
                    premState,
                    maIsSym
            );

            mRules.add(EEngine.implies(h, postStor));
        }

    }

    private void MSIZE59hornclause() {
        EExpr h, b;

        EExpr premState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                varMap.getIntExpr(EVarInt.size));*/
        if (memBound) {
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            h = premState;
            b = pred.mStatePredMem(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            varMap.getIntExpr(EVarInt.size), varMap.getIntExpr(EVarInt.aw)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            mRules.add(EEngine.implies(h, b));
        }else{
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            h = premState;
            b = pred.mStatePred(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            varMap.getIntExpr(EVarInt.size), varMap.getIntExpr(EVarInt.aw)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                    cd);

            mRules.add(EEngine.implies(h, b));
            EExpr preMem = pred.memPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            h = EEngine.and(
                    premState,
                    preMem
            );

            b = pred.memPred(contractID, next,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

            mRules.add(EEngine.implies(h, b));
        }

        EExpr pregState = pred.gStatePred(contractID, pc,
                
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);

        h = EEngine.and(premState,
                pregState);

        b = pred.gStatePred(contractID, next,
                
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));


        EExpr preStor = pred.storPred(contractID, pc,
                
                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        h = EEngine.and(premState,
                preStor);

        b = pred.storPred(contractID, next,
                
                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void JUMP56hornclause() {
        EExpr h, b;
            String jump = null;

            ArrayList<String> arg = bytecode.getComputedArguments();
            if (arg.size()>0) {
                if (arg.get(0) != null) {
                    jump = arg.get(0);
                }
            }

            Integer jumpArg;
            jumpArg = jump != null ? Integer.parseInt(jump) : null;

            if (jumpArg == null){ // when we skip pre-analysis jumpArg should be read in the lexing results
                if (!bytecode.getArg().equals("ignore")) {
                    jumpArg = Integer.parseInt(bytecode.getArg().trim(), 16);
                }
            }

            if (jumpArg != null) {
                if (jumpDests.contains(jumpArg)) {
                    EExpr premState;

                    if (memBound) {
                        premState = pred.mStatePredMem(contractID, pc,
                                varMap.getIntExpr(EVarInt.size),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                varMap.getTaintExpr(Taint.taint),
                                cd, standardMem);
                    }
                    else{
                        premState = pred.mStatePred(contractID, pc,
                                varMap.getIntExpr(EVarInt.size),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                varMap.getTaintExpr(Taint.taint),
                                cd);
                    }
                    /*EExpr valid = EEngine.checkValid(
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 1));*/
                    EExpr topOfStack = EEngine.eq(varMap.getIntExpr(EVarInt.jumpDest),
                            EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1)));

                    EExpr validJump = EEngine.eq(
                            EEngine.num(jumpArg),
                            varMap.getIntExpr(EVarInt.jumpDest)
                    );

                    if (memBound){
                        h = EEngine.and(
                                premState,
                                topOfStack,
                                validJump
                        );

                        b = pred.mStatePredMem(contractID, jumpArg,
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                varMap.getTaintExpr(Taint.taint),
                                cd, standardMem);
                        mRules.add(EEngine.implies(h, b));
                    }
                    else{
                        h = EEngine.and(
                                premState,
                                topOfStack,
                                validJump
                        );

                        b = pred.mStatePred(contractID, jumpArg,
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                varMap.getTaintExpr(Taint.taint),
                                cd);
                        mRules.add(EEngine.implies(h, b));

                        EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                        h = EEngine.and(
                                premState,
                                topOfStack,
                                validJump,
                                preMem
                        );

                        b = pred.memPred(contractID, jumpArg, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                        mRules.add(EEngine.implies(h, b));
                    }

                    EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
                    h = EEngine.and(
                            premState,
                            topOfStack,
                            validJump,
                            pregState
                    );

                    b = pred.gStatePred(contractID, jumpArg,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
                    mRules.add(EEngine.implies(h, b));

                    EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
                    h = EEngine.and(
                            premState,
                            topOfStack,
                            validJump,
                            preStor
                    );

                    b = pred.storPred(contractID, jumpArg,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
                    mRules.add(EEngine.implies(h, b));
                }
                else {
                    //invalidJump(pc);
                    if (debug) {
                        LOGGER.error("Jump to invalid destination detected at PC:" + pc);
                    }
                }
            } else {
                //jumpEverywhere(pc);
                if (debug) {
                    LOGGER.error("Jump destination unknown: cannot be found in a PUSH before PC: " + Integer.toString(pc));
                }
            }
    }

    private void JUMPI57hornclause() {
            String jump = null;
            String cond = null;

            ArrayList<String> arg = bytecode.getComputedArguments();

            if (arg.size() > 0) {
                if (arg.get(0) != null) {
                    jump = arg.get(0);
                }

                if (arg.get(1) != null) {
                    cond = arg.get(1);
                }
            }
            Integer jumpArg, condArg;
            jumpArg = jump != null ? Integer.parseInt(jump) : null;
            condArg = cond != null ? Integer.parseInt(cond) : null;

            if (jumpArg == null){ // when we skip pre-analysis jumpArg should be read in the lexing results
                if (!bytecode.getArg().equals("ignore")) {
                    jumpArg = Integer.parseInt(bytecode.getArg().trim(), 16);
                }
            }

            if (jumpArg != null) {
                if (jumpDests.contains(jumpArg)) {
                    EExpr condition = getYArg();
                    EExpr conditionT = getYArgT();
                    EExpr conditionS = getYArgS();

                    // Condition to jump

                    EExpr cJump = EEngine.ite(
                            EEngine.eq(varMap.getSymbolicExpr(Sym.ys),
                                    EEngine.bool(true)),
                            EEngine.bool(true),
                            EEngine.eq(varMap.getIntExpr(EVarInt.y),
                                    EEngine.num(1)
                            ));

                    // Condition for pc + 1
                    EExpr cPcPlus1 = EEngine.ite(EEngine.eq(varMap.getSymbolicExpr(Sym.ys),
                            EEngine.bool(true)),
                            EEngine.bool(true),
                            EEngine.eq(varMap.getIntExpr(EVarInt.y),
                                    EEngine.num(0)
                            ));

                    EExpr premState;

                    if (memBound){
                        premState = pred.mStatePredMem(contractID, pc,
                                varMap.getIntExpr(EVarInt.size),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                varMap.getTaintExpr(Taint.taint),
                                cd, standardMem);
                    }
                    else{
                        premState = pred.mStatePred(contractID, pc,
                                varMap.getIntExpr(EVarInt.size),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                varMap.getTaintExpr(Taint.taint),
                                cd);
                    }


                    /*EExpr valid = EEngine.checkValid(
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
                    );*/
                    EExpr topOfStack = EEngine.eq(varMap.getIntExpr(EVarInt.jumpDest),
                            EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1)));

                    // here the valid should not be in a loop, we encode it using ite (similar to the standard valid)
                    EExpr validJump = EEngine.eq(
                            EEngine.num(jumpArg),
                            varMap.getIntExpr(EVarInt.jumpDest)
                    );

                    EExpr checkPremsJUMPIpredicate = pred.checkPremsJUMPIPred(contractID, pc,
                            varMap.getIntExpr(EVarInt.jumpDest),
                            varMap.getIntExpr(EVarInt.y), varMap.getTaintExpr(Taint.yt), varMap.getSymbolicExpr(Sym.ys), varMap.getTaintExpr(Taint.taint), cd);

                    EExpr h;

                    h = EEngine.and(
                            premState,
                            validJump,
                            topOfStack,
                            condition,
                            conditionT,
                            conditionS
                    );

                    EExpr b = checkPremsJUMPIpredicate;
                    mRules.add(EEngine.implies(h, b));

                    if (memBound) {
                        h = EEngine.and(
                                premState,
                                validJump,
                                topOfStack,
                                condition,
                                conditionT,
                                conditionS,
                                cJump
                        );

                        b = pred.mStatePredMem(contractID, jumpArg,
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.or(varMap.getTaintExpr(Taint.taint), varMap.getTaintExpr(Taint.yt)),
                                cd, standardMem);

                        mRules.add(EEngine.implies(h, b));

                        h = EEngine.and(
                                premState,
                                validJump,
                                topOfStack,
                                condition,
                                conditionT,
                                conditionS,
                                cPcPlus1
                        );

                        b = pred.mStatePredMem(contractID, next,
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.or(varMap.getTaintExpr(Taint.taint), varMap.getTaintExpr(Taint.yt)),
                                cd, standardMem);
                        mRules.add(EEngine.implies(h, b));
                    }
                    else{

                        h = EEngine.and(
                                premState,
                                validJump,
                                topOfStack,
                                condition,
                                conditionT,
                                conditionS,
                                cJump
                        );

                        b = pred.mStatePred(contractID, jumpArg,
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.or(varMap.getTaintExpr(Taint.taint), varMap.getTaintExpr(Taint.yt)),
                                cd);

                        mRules.add(EEngine.implies(h, b));

                        h = EEngine.and(
                                premState,
                                validJump,
                                topOfStack,
                                condition,
                                conditionT,
                                conditionS,
                                cPcPlus1
                        );

                        b = pred.mStatePred(contractID, next,
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
                                varMap.getArrExpr(EVarArr.stack),
                                varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                                EEngine.or(varMap.getTaintExpr(Taint.taint), varMap.getTaintExpr(Taint.yt)),
                                cd);
                        mRules.add(EEngine.implies(h, b));

                        EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                        h = EEngine.and(
                                premState,
                                validJump,
                                topOfStack,
                                condition,
                                conditionT,
                                conditionS,
                                preMem,
                                cJump
                        );

                        b = pred.memPred(contractID, jumpArg, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                        mRules.add(EEngine.implies(h, b));

                        h = EEngine.and(
                                premState,
                                validJump,
                                topOfStack,
                                condition,
                                conditionT,
                                conditionS,
                                preMem,
                                cPcPlus1
                        );

                        b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                        mRules.add(EEngine.implies(h, b));
                    }

                    EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
                    h = EEngine.and(
                            premState,
                            validJump,
                            topOfStack,
                            condition,
                            conditionT,
                            conditionS,
                            pregState,
                            cJump
                    );

                    b = pred.gStatePred(contractID, jumpArg,
                             varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
                    mRules.add(EEngine.implies(h, b));

                    h = EEngine.and(
                            premState,
                            validJump,
                            topOfStack,
                            condition,
                            conditionT,
                            conditionS,
                            pregState,
                            cPcPlus1
                    );


                    b = pred.gStatePred(contractID, next,
                             varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
                    mRules.add(EEngine.implies(h, b));

                    EExpr preStor = pred.storPred(contractID, pc, 
                            varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

                    h = EEngine.and(
                            premState,
                            validJump,
                            topOfStack,
                            condition,
                            conditionT,
                            conditionS,
                            preStor,
                            cJump
                    );

                    b = pred.storPred(contractID, jumpArg,
                             varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
                    mRules.add(EEngine.implies(h, b));

                    h = EEngine.and(
                            premState,
                            validJump,
                            topOfStack,
                            condition,
                            conditionT,
                            conditionS,
                            preStor,
                            cPcPlus1
                    );


                    b = pred.storPred(contractID, next,
                             varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
                    mRules.add(EEngine.implies(h, b));
                }
                else {
                    //invalidJumpI(pc);
                    if (debug) {
                        LOGGER.error("Jump to invalid destination detected at PC:" + pc);
                    }
                }
            } else {
                //jumpIEverywhere(pc);
                if (debug) {
                    LOGGER.error("Jump destination unknown: cannot be found in a PUSH before PC:" + pc);
                }
            }
    }

//    public void jumpEverywhere(int pc) {
//        EExpr h, b;
//        for (int jumpArg : jumpDests) {
//            EExpr premState;
//
//            if (memBound){
//                premState = pred.mStatePredMem(contractID, pc,
//                        varMap.getIntExpr(EVarInt.size),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        varMap.getTaintExpr(Taint.taint),
//                        cd, standardMem);
//            }
//            else{
//                premState = pred.mStatePred(contractID, pc,
//                        varMap.getIntExpr(EVarInt.size),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        varMap.getTaintExpr(Taint.taint),
//                        cd);
//            }
//
//            /*EExpr valid = EEngine.checkValid(
//                    varMap.getIntExpr(EVarInt.size),
//                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 1));*/
//
//            EExpr topOfStack = EEngine.eq(varMap.getIntExpr(EVarInt.jumpDest),
//                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1)));
//
//            EExpr validJump = EEngine.eq(
//                    EEngine.num(jumpArg),
//                    varMap.getIntExpr(EVarInt.jumpDest)
//            );
//
//            h = EEngine.and(
//                    premState,
//                    topOfStack,
//                    validJump
//            );
//
//            if (memBound){
//                b = pred.mStatePredMem(contractID, jumpArg,
//                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        varMap.getTaintExpr(Taint.taint),
//                        cd, standardMem);
//                mRules.add(EEngine.implies(h, b));
//            }
//            else{
//                b = pred.mStatePred(contractID, jumpArg,
//                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        varMap.getTaintExpr(Taint.taint),
//                        cd);
//                mRules.add(EEngine.implies(h, b));
//
//                EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
//                h = EEngine.and(
//                        premState,
//                        topOfStack,
//                        validJump,
//                        preMem
//                );
//
//                b = pred.memPred(contractID, jumpArg, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM),  cd);
//                mRules.add(EEngine.implies(h, b));
//            }
//
//            EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
//            h = EEngine.and(
//                    premState,
//                    pregState,
//                    topOfStack,
//                    validJump
//            );
//
//            b = pred.gStatePred(contractID, jumpArg,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
//            mRules.add(EEngine.implies(h, b));
//            EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
//            h = EEngine.and(
//                    premState,
//                    preStor,
//                    topOfStack,
//                    validJump
//            );
//
//            b = pred.storPred(contractID, jumpArg,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
//            mRules.add(EEngine.implies(h, b));
//        }
//    }


    //TODO: on invalid jump we just stop for now
    private void invalidJumpI(int pc) {

    }

//    private void jumpIEverywhere(int pc) {
//        for (int jumpArg : jumpDests) {
//            EExpr condition = getYArg();
//            EExpr conditionT = getYArgT();
//            EExpr conditionS = getYArgS();
//
//            // Condition to jump
//
//            EExpr cJump = EEngine.ite(
//                    EEngine.eq(varMap.getSymbolicExpr(Sym.ys),
//                            EEngine.bool(true)),
//                    EEngine.bool(true),
//                    EEngine.eq(varMap.getIntExpr(EVarInt.y),
//                            EEngine.num(1)
//                    ));
//
//            // Condition for pc + 1
//            EExpr cPcPlus1 = EEngine.ite(EEngine.eq(varMap.getSymbolicExpr(Sym.ys),
//                    EEngine.bool(true)),
//                    EEngine.bool(true),
//                    EEngine.eq(varMap.getIntExpr(EVarInt.y),
//                            EEngine.num(0)
//                    ));
//
//            EExpr premState;
//
//            if (memBound) {
//                premState = pred.mStatePredMem(contractID, pc,
//                        varMap.getIntExpr(EVarInt.size),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        varMap.getTaintExpr(Taint.taint),
//                        cd, standardMem);
//            } else {
//                premState = pred.mStatePred(contractID, pc,
//                        varMap.getIntExpr(EVarInt.size),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        varMap.getTaintExpr(Taint.taint),
//                        cd);
//            }
//
//
//            /*EExpr valid = EEngine.checkValid(
//                    varMap.getIntExpr(EVarInt.size),
//                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 2)
//            );*/
//            EExpr topOfStack = EEngine.eq(varMap.getIntExpr(EVarInt.jumpDest),
//                    EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1)));
//
//            // here the valid should not be in a loop, we encode it using ite (similar to the standard valid)
//            EExpr validJump = EEngine.eq(
//                    EEngine.num(jumpArg),
//                    varMap.getIntExpr(EVarInt.jumpDest)
//            );
//
//            EExpr checkPremsJUMPIpredicate = pred.checkPremsJUMPIPred(contractID, pc,
//                    varMap.getIntExpr(EVarInt.jumpDest),
//                    varMap.getIntExpr(EVarInt.y), varMap.getTaintExpr(Taint.yt), varMap.getSymbolicExpr(Sym.ys), varMap.getTaintExpr(Taint.taint), cd);
//
//            EExpr h;
//
//            h = EEngine.and(
//                    premState,
//                    validJump,
//                    topOfStack,
//                    condition,
//                    conditionT,
//                    conditionS
//            );
//
//            EExpr b = checkPremsJUMPIpredicate;
//            mRules.add(EEngine.implies(h, b));
//
//            if (memBound) {
//                h = EEngine.and(
//                        premState,
//                        checkPremsJUMPIpredicate,
//                        cJump
//                );
//
//                b = pred.mStatePredMem(contractID, jumpArg,
//                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        EEngine.or(varMap.getTaintExpr(Taint.taint), varMap.getTaintExpr(Taint.yt)),
//                        cd, standardMem);
//
//                mRules.add(EEngine.implies(h, b));
//
//                h = EEngine.and(
//                        premState,
//                        checkPremsJUMPIpredicate,
//                        cPcPlus1
//                );
//
//                b = pred.mStatePredMem(contractID, next,
//                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        EEngine.or(varMap.getTaintExpr(Taint.taint), varMap.getTaintExpr(Taint.yt)),
//                        cd, standardMem);
//                mRules.add(EEngine.implies(h, b));
//            } else {
//
//                h = EEngine.and(
//                        premState,
//                        checkPremsJUMPIpredicate,
//                        cJump
//                );
//
//                b = pred.mStatePred(contractID, jumpArg,
//                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        EEngine.or(varMap.getTaintExpr(Taint.taint), varMap.getTaintExpr(Taint.yt)),
//                        cd);
//
//                mRules.add(EEngine.implies(h, b));
//
//                h = EEngine.and(
//                        premState,
//                        checkPremsJUMPIpredicate,
//                        cPcPlus1
//                );
//
//                b = pred.mStatePred(contractID, next,
//                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 2),
//                        varMap.getArrExpr(EVarArr.stack),
//                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
//                        EEngine.or(varMap.getTaintExpr(Taint.taint), varMap.getTaintExpr(Taint.yt)),
//                        cd);
//                mRules.add(EEngine.implies(h, b));
//
//                EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
//                h = EEngine.and(
//                        premState,
//                        checkPremsJUMPIpredicate,
//                        preMem,
//                        cJump
//                );
//
//                b = pred.memPred(contractID, jumpArg, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
//                mRules.add(EEngine.implies(h, b));
//
//                h = EEngine.and(
//                        premState,
//                        checkPremsJUMPIpredicate,
//                        preMem,
//                        cPcPlus1
//                );
//
//                b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
//                mRules.add(EEngine.implies(h, b));
//            }
//
//            EExpr pregState = pred.gStatePred(contractID, pc, varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
//            h = EEngine.and(
//                    premState,
//                    checkPremsJUMPIpredicate,
//                    pregState,
//                    cJump
//            );
//
//            b = pred.gStatePred(contractID, jumpArg,
//                    varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
//            mRules.add(EEngine.implies(h, b));
//
//            h = EEngine.and(
//                    premState,
//                    checkPremsJUMPIpredicate,
//                    pregState,
//                    cPcPlus1
//            );
//
//
//            b = pred.gStatePred(contractID, next,
//                    varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
//            mRules.add(EEngine.implies(h, b));
//
//            EExpr preStor = pred.storPred(contractID, pc,
//                    varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
//
//            h = EEngine.and(
//                    premState,
//                    checkPremsJUMPIpredicate,
//                    preStor,
//                    cJump
//            );
//
//            b = pred.storPred(contractID, jumpArg,
//                    varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
//            mRules.add(EEngine.implies(h, b));
//
//            h = EEngine.and(
//                    premState,
//                    checkPremsJUMPIpredicate,
//                    preStor,
//                    cPcPlus1
//            );
//
//
//            b = pred.storPred(contractID, next,
//                    varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
//            mRules.add(EEngine.implies(h, b));
//        }
//    }

    private void JUMPDEST5bhornclause() {
        EExpr h, b;

        EExpr mState;
        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                varMap.getIntExpr(EVarInt.size)
        );*/
        if (memBound) {
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            h = mState;

            b = pred.mStatePredMem(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            mRules.add(EEngine.implies(h, b));
        }
        else{
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            h = mState;

            b = pred.mStatePred(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            mRules.add(EEngine.implies(h, b));

            // Propagation of remaining memory predicates

            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            h = EEngine.and(
                    mState,
                    preMem
            );
            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            mRules.add(EEngine.implies(h, b));
        }
        //Propagation of gstate and storage

        EExpr gState = pred.gStatePred(contractID, pc,   varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                mState,
                gState
        );
        b = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        h = EEngine.and(
                mState,
                preStor
        );
        b = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void PUSH160hornclause(Integer x) {
        EExpr h, b;
        BigInteger arg = new BigInteger(bytecode.getArg(), 16);

        EExpr push1arg = EEngine.num(arg.toString());
        EExpr premState;
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                varMap.getIntExpr(EVarInt.size));*/
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            h = premState;

            b = pred.mStatePredMem(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            varMap.getIntExpr(EVarInt.size), push1arg),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);

            mRules.add(EEngine.implies(h, b));
        } else {
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);

            h = premState;

            b = pred.mStatePred(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack),
                            varMap.getIntExpr(EVarInt.size), push1arg),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                    cd);

            mRules.add(EEngine.implies(h, b));
            EExpr preMem = pred.memPred(contractID, pc,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            h = EEngine.and(
                    premState,
                    preMem
            );

            b = pred.memPred(contractID, next,
                    varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);

            mRules.add(EEngine.implies(h, b));
        }

        EExpr pregState = pred.gStatePred(contractID, pc,
                
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);

        h = EEngine.and(premState, pregState);

        b = pred.gStatePred(contractID, next,
                
                varMap.getIntExpr(EVarInt.n),
                varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));


        EExpr preStor = pred.storPred(contractID, pc,
                
                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);

        h = EEngine.and(premState,
                preStor
        );

        b = pred.storPred(contractID, next,
                
                varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void DUPS80hornclause(Integer x) {
        EExpr premState;
        /*EExpr valid =
                EEngine.checkValid(
                        EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x))
                );*/
        if (memBound){
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }


        EExpr val = EEngine.eq(
                varMap.getIntExpr(EVarInt.val0),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x)))
        );
        EExpr valT = EEngine.eq(
                varMap.getTaintExpr(Taint.xt),
                EEngine.selectT(varMap.getArrTExpr(EVarArrT.stackT),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x)))
        );
        EExpr valS = EEngine.eq(
                varMap.getSymbolicExpr(Sym.xs),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x)))
        );

        EExpr postmState;
        if (memBound){
            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), varMap.getIntExpr(EVarInt.size),
                            varMap.getIntExpr(EVarInt.val0)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            varMap.getTaintExpr(Taint.xt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            varMap.getSymbolicExpr(Sym.xs)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            postmState = pred.mStatePred(contractID, next,
                    EEngine.add(varMap.getIntExpr(EVarInt.size), 1),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), varMap.getIntExpr(EVarInt.size),
                            varMap.getIntExpr(EVarInt.val0)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            varMap.getIntExpr(EVarInt.size),
                            varMap.getTaintExpr(Taint.xt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getIntExpr(EVarInt.size),
                            varMap.getSymbolicExpr(Sym.xs)), varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem),
                    varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(premState,
                    preMem), postMem);
            mRules.add(rule2);
        }
        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor),
                varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(EEngine.and(premState, val, valT, valS), postmState);
        EExpr rule3 = EEngine.implies(
                EEngine.and(
                    premState,
                    pregState
                ),
                postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void SWAPS90hornclause(Integer x) {
        EExpr premState;

        if (memBound){
            premState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }


        EExpr v1 = EEngine.eq(
                varMap.getIntExpr(EVarInt.val0),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        EExpr v1T = EEngine.eq(
                varMap.getTaintExpr(Taint.xt),
                EEngine.selectT(varMap.getArrTExpr(EVarArrT.stackT),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        EExpr v1S = EEngine.eq(
                varMap.getSymbolicExpr(Sym.xs),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), 1))
        );

        EExpr v2 =
                EEngine.eq(
                        varMap.getIntExpr(EVarInt.val1),
                        EEngine.selectInt(varMap.getArrExpr(EVarArr.stack),
                                EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)))
                );

        EExpr v2T = EEngine.eq(
                varMap.getTaintExpr(Taint.yt),
                EEngine.selectT(varMap.getArrTExpr(EVarArrT.stackT),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)))
        );

        EExpr v2S = EEngine.eq(
                varMap.getSymbolicExpr(Sym.ys),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS),
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)))
        );

        EExpr store1 =
                EEngine.eq(
                        varMap.getArrExpr(EVarArr.newStack),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                varMap.getIntExpr(EVarInt.val1)
                        )
                );
        EExpr store1T =
                EEngine.eq(
                        varMap.getArrTExpr(EVarArrT.newStackT),
                        EEngine.storeT(varMap.getArrTExpr(EVarArrT.stackT), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                varMap.getTaintExpr(Taint.yt)
                        )
                );
        EExpr store1S =
                EEngine.eq(
                        varMap.getArrSExpr(EVarArrS.newStackS),
                        EEngine.storeS(varMap.getArrSExpr(EVarArrS.stackS), EEngine.sub(varMap.getIntExpr(EVarInt.size), 1),
                                varMap.getSymbolicExpr(Sym.ys)
                        )
                );



        /*EExpr valid = EEngine.checkValid(
                varMap.getIntExpr(EVarInt.size),
                EEngine.sub(varMap.getIntExpr(EVarInt.size),
                EEngine.num(x + 1)));*/

        EExpr h = EEngine.and(
                premState,
                v1,
                v2,
                store1,
                store1S,
                store1T,
                v1T,
                v1S,
                v2T,
                v2S
        );

        EExpr b;
        if (memBound){
            b = pred.mStatePredMem(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.newStack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)),
                            varMap.getIntExpr(EVarInt.val0)
                    ),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)),
                            varMap.getTaintExpr(Taint.xt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)),
                            varMap.getSymbolicExpr(Sym.xs)), varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            mRules.add(
                    EEngine.implies(h, b)
            );
        }
        else{
            b = pred.mStatePred(contractID, next,
                    varMap.getIntExpr(EVarInt.size),
                    EEngine.storeInt(
                            varMap.getArrExpr(EVarArr.newStack),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)),
                            varMap.getIntExpr(EVarInt.val0))
                    ,
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)),
                            varMap.getTaintExpr(Taint.xt)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(x + 1)),
                            varMap.getSymbolicExpr(Sym.xs)), varMap.getTaintExpr(Taint.taint),
                    cd);
            mRules.add(
                    EEngine.implies(h, b)
            );
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va),
                    varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va),
                    varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(premState, preMem), postMem);
            mRules.add(rule2);
        }

        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor),
                varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor),
                varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule3 = EEngine.implies(EEngine.and(premState, pregState), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState), postStor);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void REVERT(){

    }

    private void dummy(int pop, Integer push, boolean pushes) {
        EExpr h, b;
        int sub;
        EExpr pushedVal;
        if (pushes) {
            if (push != null) {
                sub = pop - 1;
                pushedVal = EEngine.num(push);
            } else {
                sub = pop;
                pushedVal = varMap.getIntExpr(EVarInt.val0);
            }
        }
        else {
            sub = pop;
            pushedVal = null;
        }

        EExpr mState;
        if (memBound) {
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }

       /* EExpr valid = EEngine.checkValid(
                EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(sub)),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(pop)));*/
        h = mState;

        if (pushes){
            if (memBound) {
                b = pred.mStatePredMem(contractID, next,
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(sub)),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), varMap.getIntExpr(EVarInt.size), pushedVal),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                varMap.getIntExpr(EVarInt.size),
                                EEngine.bool(true)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                varMap.getIntExpr(EVarInt.size),
                                EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                        cd, standardMem);
            }
            else{
                b = pred.mStatePred(contractID, next,
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(sub)),
                        EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), varMap.getIntExpr(EVarInt.size), pushedVal),
                        EEngine.storeT(
                                varMap.getArrTExpr(EVarArrT.stackT),
                                varMap.getIntExpr(EVarInt.size),
                                EEngine.bool(true)),
                        EEngine.storeS(
                                varMap.getArrSExpr(EVarArrS.stackS),
                                varMap.getIntExpr(EVarInt.size),
                                EEngine.bool(true)), varMap.getTaintExpr(Taint.taint),
                        cd);
            }
        } else {
            if (memBound) {
                b = pred.mStatePredMem(contractID, next,
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(sub)),
                        varMap.getArrExpr(EVarArr.stack),
                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd,
                        standardMem);
                mRules.add(EEngine.implies(h, b));

            } else {
                b = pred.mStatePred(contractID, next,
                        EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(sub)),
                        varMap.getArrExpr(EVarArr.stack),
                        varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                        varMap.getTaintExpr(Taint.taint),
                        cd);
                mRules.add(EEngine.implies(h, b));

                // Propagation of remaining memory predicates

                EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                h = EEngine.and(
                        mState,
                        preMem
                );
                b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.x), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
                mRules.add(EEngine.implies(h, b));
            }

        }
        //Propagation of gstate and storage

        EExpr gState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        h = EEngine.and(
                mState,
                gState
        );
        b = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        h = EEngine.and(
                mState,
                preStor
        );
        b = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        mRules.add(EEngine.implies(h, b));
    }

    private void LOGSa0hornclause(Integer x) {
        JUMPDEST5bhornclause();
    }

    private void NOOPffhornclause() {
        LOGGER.error("SKIPPING: " + pc);
        //EExpr valid = EEngine.checkValid(varMap.getIntExpr(EVarInt.size), varMap.getIntExpr(EVarInt.size));
        Integer opCode = bytecode.getOpCode();
        Integer deleteFromStack = 0;
        if (opCode == 22) { //and
            deleteFromStack = 1;
        }
        if (opCode == 23) { //or
            deleteFromStack = 1;
        }
        if (opCode == 57) { //codecopy
            deleteFromStack = 3;

        }
        if (opCode == 81) { //mload
            deleteFromStack = 0;
        }
        if (opCode == 82) { //mstore
            deleteFromStack = 2;
        }
        EExpr premState, postmState;

        if (memBound){
            premState = pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
            postmState = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(deleteFromStack)), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            premState = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            postmState = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(deleteFromStack)), varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
            EExpr preMem = pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr postMem = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.va), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            EExpr rule2 = EEngine.implies(EEngine.and(preMem, premState), postMem);
            mRules.add(rule2);
        }

        EExpr pregState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr postgState = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        EExpr preStor = pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr postStor = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd);
        EExpr rule1 = EEngine.implies(premState, postmState);
        EExpr rule3 = EEngine.implies(EEngine.and(pregState, premState), postgState);
        EExpr rule4 = EEngine.implies(EEngine.and(preStor, premState), postStor);
        mRules.add(rule1);
        mRules.add(rule3);
        mRules.add(rule4);
    }

    private void RETURN() {

    }

    private void SELFDESTRUCT() {

    }

    private void CREATE() {
        CALL();
    }

    private void CALLCODE() {
        CALL();
    }

    private void DELEGATECALL() {
        CALL();
    }

    private void addQueryRe(){
        EExpr query;
        if (memBound){
            query = EEngine.and(
                    pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                            varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getTaintExpr(Taint.taint),
                    1, varMap.getMem())
            );
        }
        else{
            query = pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    1
            );
        }
        LOGGER.info("Adding a reentrancy check at pc " + pc);
        mQueriesRe.add(query);
    }

    private void addQueryDe(){
        EExpr query;
        if (memBound){
            query = EEngine.and(pred.mStatePredMem(contractID, pc, varMap.getIntExpr(EVarInt.size),
                            varMap.getArrExpr(EVarArr.stack),
                            varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                            varMap.getTaintExpr(Taint.taint),
                            0, varMap.getMem()),
                    EEngine.or(EEngine.selectT(varMap.getArrTExpr(EVarArrT.stackT), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(3))),
                            varMap.getTaintExpr(Taint.taint)
                            )
            );
        }
        else{
            query = EEngine.and(pred.mStatePred(contractID, pc, varMap.getIntExpr(EVarInt.size),
                    varMap.getArrExpr(EVarArr.stack),
                    varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    0),
                    EEngine.or(
                            EEngine.selectT(varMap.getArrTExpr(EVarArrT.stackT), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(3))),
                            varMap.getTaintExpr(Taint.taint)
                            )
            );
        }
        LOGGER.info("Adding a dependency check at pc " + pc);
        mQueriesDe.add(query);
    }

    private void CALL(){
        //Callee is unknown
        EExpr h, b;

        EExpr mState;

        if (memBound){
            mState = pred.mStatePredMem(contractID, pc,
                    varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd, standardMem);
        }
        else{
            mState = pred.mStatePred(contractID, pc,
                    varMap.getIntExpr(EVarInt.size), varMap.getArrExpr(EVarArr.stack), varMap.getArrTExpr(EVarArrT.stackT), varMap.getArrSExpr(EVarArrS.stackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd);
        }
        EExpr gasTo = EEngine.eq(
                varMap.getIntExpr(EVarInt.gasTo),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(1)))
        );
        EExpr to = EEngine.eq(
                varMap.getIntExpr(EVarInt.to),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(2)))
        );
        EExpr va = EEngine.eq(
                varMap.getIntExpr(EVarInt.va),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(3)))
        );
        EExpr io = EEngine.eq(
                varMap.getIntExpr(EVarInt.io),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(4)))
        );
        EExpr is = EEngine.eq(
                varMap.getIntExpr(EVarInt.is),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(5)))
        );
        EExpr oo = EEngine.eq(
                varMap.getIntExpr(EVarInt.oo),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(6)))
        );
        EExpr os = EEngine.eq(
                varMap.getIntExpr(EVarInt.os),
                EEngine.selectInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(7)))
        );
        /*EExpr valid = EEngine.checkValid(
                EEngine.add(varMap.getIntExpr(EVarInt.size), EEngine.num(1)),
                EEngine.sub(varMap.getIntExpr(EVarInt.size), 7));*/
       /* EExpr exenv = pred.exenvPred(contractID,
                
                varMap.getIntExpr(EVarInt.i),
                cd,
                varMap.getIntExpr(EVarInt.v), varMap.getIntExpr(EVarInt.inputsize));*/
        EExpr gState = pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.val0), varMap.getIntExpr(EVarInt.val1), cd);

        EExpr vas = EEngine.eq(
                varMap.getSymbolicExpr(Sym.zs),
                EEngine.selectS(varMap.getArrSExpr(EVarArrS.stackS), EEngine.sub(varMap.getIntExpr(EVarInt.size), EEngine.num(3)))
        );

        EExpr vaLEb = EEngine.ite(varMap.getSymbolicExpr(Sym.zs),
                EEngine.bool(true),
                EEngine.le(
                    varMap.getIntExpr(EVarInt.va),
                    varMap.getIntExpr(EVarInt.val1)
                )
        );

        h = EEngine.and(
                mState,
                gasTo,
                to,
                va,
                io,
                is,
                oo,
                os
                //valid
                //gState
                //vas,
                //vaLEb
        );

        EExpr checkPremsUnknown = pred.checkPremsCALLunknownPred(contractID, pc,
                varMap.getIntExpr(EVarInt.to),
                varMap.getIntExpr(EVarInt.va),
                varMap.getIntExpr(EVarInt.oo),
                varMap.getIntExpr(EVarInt.os),
                varMap.getTaintExpr(Taint.taint),
                cd
        );

        b = checkPremsUnknown;

        mRules.add(EEngine.implies(h, b));

        if (memBound){
            h = EEngine.and(
                    mState,
                    gasTo,
                    to,
                    va,
                    io,
                    is,
                    oo,
                    os
            );

            b = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 6),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.num(1)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                    cd, ovappxMem);
            mRules.add(EEngine.implies(h, b));

            b = pred.mStatePredMem(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 6),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.num(0)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                    cd, ovappxMem);
            mRules.add(EEngine.implies(h, b));

        }
        else{
            h = EEngine.and(
                    mState,
                    gasTo,
                    to,
                    va,
                    io,
                    is,
                    oo,
                    os
            );

            b = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 6),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.num(1)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                    cd);
            mRules.add(EEngine.implies(h, b));

            b = pred.mStatePred(contractID, next,
                    EEngine.sub(varMap.getIntExpr(EVarInt.size), 6),
                    EEngine.storeInt(varMap.getArrExpr(EVarArr.stack), EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.num(0)),
                    EEngine.storeT(
                            varMap.getArrTExpr(EVarArrT.stackT),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.bool(false)),
                    EEngine.storeS(
                            varMap.getArrSExpr(EVarArrS.stackS),
                            EEngine.sub(varMap.getIntExpr(EVarInt.size), 7),
                            EEngine.bool(false)), varMap.getTaintExpr(Taint.taint),
                    cd);
            mRules.add(EEngine.implies(h, b));

            //Arbitrary values can be written to the specified memory fragment

            h = EEngine.and(
                    checkPremsUnknown,
                    EEngine.ge(varMap.getIntExpr(EVarInt.pos), varMap.getIntExpr(EVarInt.oo)),
                    EEngine.lt(varMap.getIntExpr(EVarInt.pos), EEngine.add(varMap.getIntExpr(EVarInt.oo), varMap.getIntExpr(EVarInt.os))),
                    EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.pos), EEngine.num(32)), EEngine.num(0))
                    );
            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.posmem), varMap.getIntExpr(EVarInt.v), EEngine.bool(true), EEngine.bool(true), cd);
            mRules.add(EEngine.implies(h, b));

            //All other values are propagated

            h = EEngine.and(
                    checkPremsUnknown,
                    EEngine.lt(varMap.getIntExpr(EVarInt.pos), varMap.getIntExpr(EVarInt.oo)),
                    EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.pos), EEngine.num(32)), EEngine.num(0)),
                    pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.pos), varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd)
            );
            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.pos), varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            mRules.add(EEngine.implies(h, b));

            h = EEngine.and(
                    checkPremsUnknown,
                    EEngine.ge(varMap.getIntExpr(EVarInt.pos), EEngine.add(varMap.getIntExpr(EVarInt.oo), varMap.getIntExpr(EVarInt.os))),
                    EEngine.eq(EEngine.mod(varMap.getIntExpr(EVarInt.pos), EEngine.num(32)), EEngine.num(0)),
                    pred.memPred(contractID, pc, varMap.getIntExpr(EVarInt.pos), varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd)
            );
            b = pred.memPred(contractID, next, varMap.getIntExpr(EVarInt.pos), varMap.getIntExpr(EVarInt.v), varMap.getTaintExpr(Taint.taintM), varMap.getSymbolicExpr(Sym.symM), cd);
            mRules.add(EEngine.implies(h, b));
        }

        //After execution the global state might have arbitrary values - if call successful
        h = checkPremsUnknown;
        b = pred.storPred(contractID, next,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), EEngine.bool(true), EEngine.bool(true), cd);
        mRules.add(EEngine.implies(h, b));

        h = checkPremsUnknown;
        b = pred.gStatePred(contractID, next,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd);
        mRules.add(EEngine.implies(h, b));

        //The execution of the contract might be reentered at an arbitrary (higher) call level

        if (memBound){
            h = checkPremsUnknown;
            b = pred.mStatePredMem(contractID,
                    0,
                    EEngine.num(0),
                    varMap.getArrExpr(EVarArr.initstack),
                    varMap.getArrTExpr(EVarArrT.initstackT),
                    varMap.getArrSExpr(EVarArrS.initstackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd + 1, varMap.getMemZero());
            mRules.add(EEngine.implies(h, b));

        }
        else{
            h = checkPremsUnknown;
            b = pred.mStatePred(contractID,
                    0,
                    EEngine.num(0),
                    varMap.getArrExpr(EVarArr.initstack),
                    varMap.getArrTExpr(EVarArrT.initstackT),
                    varMap.getArrSExpr(EVarArrS.initstackS),
                    varMap.getTaintExpr(Taint.taint),
                    cd + 1);
            mRules.add(EEngine.implies(h, b));

            b = pred.memPred(contractID, 0, varMap.getIntExpr(EVarInt.pos), EEngine.num(0), EEngine.bool(false), EEngine.bool(false), cd + 1);
            mRules.add(EEngine.implies(h, b));
        }

        /*h = checkPremsUnknown;

        b = pred.exenvPred(contractID,
                varMap.getIntExpr(EVarInt.x),
                varMap.getIntExpr(EVarInt.y),
                cd + 1,
                varMap.getIntExpr(EVarInt.val0),
                varMap.getIntExpr(EVarInt.val1));
        mRules.add(EEngine.implies(h, b));

        b = pred.inputDataPred(contractID,
                varMap.getIntExpr(EVarInt.val0),
                varMap.getIntExpr(EVarInt.val1),
                cd + 1
        );
        mRules.add(EEngine.implies(h, b));*/

        h = EEngine.and(
                checkPremsUnknown,
                pred.gStatePred(contractID, pc,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd)
        );

        b = pred.gStatePred(contractID, 0,  varMap.getIntExpr(EVarInt.val0), varMap.getIntExpr(EVarInt.val1), cd + 1);
        mRules.add(EEngine.implies(h, b));

//        h = EEngine.and(
//                checkPremsUnknown,
//                pred.gStatePred(contractID, pc, varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd),
//                EEngine.not(EEngine.eq(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.a)))
//        );
//
//        b = pred.gStatePred(contractID, 0,  varMap.getIntExpr(EVarInt.n), varMap.getIntExpr(EVarInt.b), cd + 1);
//        mRules.add(EEngine.implies(h, b));

        h = EEngine.and(
                checkPremsUnknown,
                pred.storPred(contractID, pc,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd)
        );

        b = pred.storPred(contractID, 0,  varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS),cd + 1);
        mRules.add(EEngine.implies(h, b));

        /*h = EEngine.and(
                checkPremsUnknown,
                pred.storPred(contractID, pc, varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd),
                EEngine.not(EEngine.eq(varMap.getIntExpr(EVarInt.x), varMap.getIntExpr(EVarInt.a)))
        );

        b = pred.storPred(contractID, 0, varMap.getIntExpr(EVarInt.posStor), varMap.getIntExpr(EVarInt.vaStor), varMap.getTaintExpr(Taint.taintS), varMap.getSymbolicExpr(Sym.symS), cd + 1);
        mRules.add(EEngine.implies(h, b));*/

        if (addQueries) {
            if (reentrCheck) {
                addQueryRe();
            }
            if (depCheck) {
                addQueryDe();
            }
        }
    }

    private void caller(){
        switch (bytecode.getOpCode()) {
            case 0:
                STOP0hornclause();
                break;
            case 1:
                ADD1hornclause();
                break;
            case 2:
                MUL2hornclause();
                break;
            case 3:
                SUB3hornclause();
                break;
            case 4:
                DIV4hornclause();
                break;
            case 5:
                SDIV5hornclause(); //TODO: over-approximated
                break;
            case 6:
                MOD6hornclause();
                break;
            case 7:
                SMOD7hornclause(); //TODO: over-approximated
                break;
            case 8:
                ADDMOD8hornclause();
                break;
            case 9:
                MULMOD9hornclause();
                break;
            case 10:
                EXP0ahornclause();
                break;
            case 11:
                SIGNEDEXTEND0bhornclause(); //TODO: over-approximated
                break;
            case 16:
                LT10hornclause();
                break;
            case 17:
                GT11hornclause();
                break;
            case 18:
                SLT12hornclause(); //TODO: over-approximated
                break;
            case 19:
                SGT13hornclause(); //TODO: over-approximated
                break;
            case 20:
                EQ14hornclause();
                break;
            case 21:
                ISZERO15hornclause();
                break;
            case 22:
                AND16hornclause();
                break;
            case 23:
                OR17hornclause();
                break;
            case 24:
                XOR18hornclause();
                break;
            case 25:
                NOT19hornclause();
                break;
            case 26:
                BYTE1ahornclause(); //TODO: over-approximated
                break;
            case 32:
                SHA320hornclause(); //TODO: over-approximated
                break;
            case 48:
                ADDRESS30hornclause(); //TODO: over-approximated
                break;
            case 49:
                BALANCE31hornclause(); //TODO: over-approximated
                break;
            case 50:
                ORIGIN31hornclause(); //TODO: over-approximated
                break;
            case 51:
                CALLER33hornclause(); //TODO: over-approximated
                break;
            case 52:
                CALLVALUE34hornclause(); //TODO: over-approximated
                break;
            case 53:
                CALLDATALOAD35hornclause();
                break;
            case 54:
                CALLDATASIZE36GAShornclause(); //TODO: over-approximated
                break;
            case 55:
                CALLDATACOPY37hornclause(); //TODO: over-approximated
                break;
            case 56:
                //dummy(0, 27, 2, 3, true);
                CODESIZE38hornclause();
                break;
            case 57:
                //dummy(3, null, 1, 3, false);
                CODECOPY39hornclause();
                break;
            case 58:
                GASPRICE3ahornclause(); //TODO: over-approximated
                break;
            case 59:
                EXTCODESIZE3bhornclause(); //TODO: over-approximated
                break;
            case 60:
                EXTCODECOPY3chornclause(); //TODO: over-approximated
                break;
            case 64:
                BLOCKHASH40hornclause(); //TODO: over-approximated
                break;
            case 65:
                COINBASE41hornclause(); //TODO: over-approximated
                break;
            case 66:
                TIMESTAMP42hornclause(); //TODO: over-approximated
                break;
            case 67:
                NUMBER43hornclause(); //TODO: over-approximated
                break;
            case 68:
                DIFFICULTY44hornclause(); //TODO: over-approximated
                break;
            case 69:
                GASLIMIT45hornclause(); //TODO: over-approximated
                break;
            case 80:
                POP50hornclause();
                break;
            case 81:
                //dummy(1, 64, 1, 3, true);
                MLOAD51hornclause();
                break;
            case 82:
                //dummy(2, null, 1, 3, false);
                MSTORE52hornclause();
                break;
            case 83:
                //dummy(2, null, 1, 3, false);
                MSTORE853hornclause();
                break;
            case 84:
                SLOAD54hornclause();
                break;
            case 85:
                SSTORE55hornclause();
                break;
            case 86:
                JUMP56hornclause();
                break;
            case 87:
                JUMPI57hornclause();
                break;
            case 88:
                PC58hornclause();
                break;
            case 89:
                MSIZE59hornclause();
                break;
            case 90:
                CALLDATASIZE36GAShornclause();
                break;
            case 91:
                JUMPDEST5bhornclause();
                break;
            case 96:
                PUSH160hornclause(1);
                break;
            case 97:
                PUSH160hornclause(2);
                break;
            case 98:
                PUSH160hornclause(3);
                break;
            case 99:
                PUSH160hornclause(4);
                break;
            case 100:
                PUSH160hornclause(5);
                break;
            case 101:
                PUSH160hornclause(6);
                break;
            case 102:
                PUSH160hornclause(7);
                break;
            case 103:
                PUSH160hornclause(8);
                break;
            case 104:
                PUSH160hornclause(9);
                break;
            case 105:
                PUSH160hornclause(10);
                break;
            case 106:
                PUSH160hornclause(11);
                break;
            case 107:
                PUSH160hornclause(12);
                break;
            case 108:
                PUSH160hornclause(13);
                break;
            case 109:
                PUSH160hornclause(14);
                break;
            case 110:
                PUSH160hornclause(15);
                break;
            case 111:
                PUSH160hornclause(16);
                break;
            case 112:
                PUSH160hornclause(17);
                break;
            case 113:
                PUSH160hornclause(18);
                break;
            case 114:
                PUSH160hornclause(19);
                break;
            case 115:
                PUSH160hornclause(20);
                break;
            case 116:
                PUSH160hornclause(21);
                break;
            case 117:
                PUSH160hornclause(22);
                break;
            case 118:
                PUSH160hornclause(23);
                break;
            case 119:
                PUSH160hornclause(24);
                break;
            case 120:
                PUSH160hornclause(25);
                break;
            case 121:
                PUSH160hornclause(26);
                break;
            case 122:
                PUSH160hornclause(27);
                break;
            case 123:
                PUSH160hornclause(28);
                break;
            case 124:
                PUSH160hornclause(29);
                break;
            case 125:
                PUSH160hornclause(30);
                break;
            case 126:
                PUSH160hornclause(31);
                break;
            case 127:
                PUSH160hornclause(32);
                break;
            case 128:
                DUPS80hornclause(1);
                break;
            case 129:
                DUPS80hornclause(2);
                break;
            case 130:
                DUPS80hornclause(3);
                break;
            case 131:
                DUPS80hornclause(4);
                break;
            case 132:
                DUPS80hornclause(5);
                break;
            case 133:
                DUPS80hornclause(6);
                break;
            case 134:
                DUPS80hornclause(7);
                break;
            case 135:
                DUPS80hornclause(8);
                break;
            case 136:
                DUPS80hornclause(9);
                break;
            case 137:
                DUPS80hornclause(10);
                break;
            case 138:
                DUPS80hornclause(11);
                break;
            case 139:
                DUPS80hornclause(12);
                break;
            case 140:
                DUPS80hornclause(13);
                break;
            case 141:
                DUPS80hornclause(14);
                break;
            case 142:
                DUPS80hornclause(15);
                break;
            case 143:
                DUPS80hornclause(16);
                break;
            case 144:
                SWAPS90hornclause(1);
                break;
            case 145:
                SWAPS90hornclause(2);
                break;
            case 146:
                SWAPS90hornclause(3);
                break;
            case 147:
                SWAPS90hornclause(4);
                break;
            case 148:
                SWAPS90hornclause(5);
                break;
            case 149:
                SWAPS90hornclause(6);
                break;
            case 150:
                SWAPS90hornclause(7);
                break;
            case 151:
                SWAPS90hornclause(8);
                break;
            case 152:
                SWAPS90hornclause(9);
                break;
            case 153:
                SWAPS90hornclause(10);
                break;
            case 154:
                SWAPS90hornclause(11);
                break;
            case 155:
                SWAPS90hornclause(12);
                break;
            case 156:
                SWAPS90hornclause(13);
                break;
            case 157:
                SWAPS90hornclause(14);
                break;
            case 158:
                SWAPS90hornclause(15);
                break;
            case 159:
                SWAPS90hornclause(16);
                break;
            case 160:
                LOGSa0hornclause(0);
                break;
            case 161:
                LOGSa0hornclause(1);
                break;
            case 162:
                LOGSa0hornclause(2);
                break;
            case 163:
                LOGSa0hornclause(3);
                break;
            case 164:
                LOGSa0hornclause(4);
                break;
            case 240:
                CREATE();
                break;
            case 241:
                CALL();
                break;
            case 242:
                CALLCODE();
                break;
            case 243:
                RETURN();
                break;
            case 244:
                DELEGATECALL();
                break;
            case 253:
                REVERT();
                break;
            //TODO: create special cases for invalid and revert
            case 254:
                REVERT(); //INVALID
                break;
            case 255:
                SELFDESTRUCT();
                break;
            default:
                NOOPffhornclause();
                break;
        }
    }
}